"""
Internal module which defines several wrapper classes for Hoc entities.
They are then available as singletons objects in neurodamus.core package
"""
from __future__ import absolute_import
import logging
import os.path
from contextlib import contextmanager
from .configuration import NeuronStdrunDefaults
from .configuration import GlobalConfig
from .configuration import SimConfig
from ..utils import classproperty


#
# Singleton, instantiated right below
#
class _Neuron(object):
    """A wrapper over the neuron simulator.
    """
    __name__ = "_Neuron"
    _h = None  # We dont import it at module-level to avoid starting neuron
    _hocs_loaded = set()

    # No new attributes. __setattr__ can rely on it
    __slots__ = ()

    @classproperty
    def h(cls):
        """The neuron hoc interpreter, initializing if needed.
        """
        return cls._h or cls._init()

    @classmethod
    def _init(cls, mpi=False):
        """Initializes the Neuron simulator.
        """
        if cls._h is not None:
            return cls._h
        if mpi:
            GlobalConfig.set_mpi()

        from neuron import h
        from neuron import nrn
        cls.__cache = {}
        cls._h = h
        cls.Section = nrn.Section
        cls.Segment = nrn.Segment
        h.load_file("stdrun.hoc")
        h("objref nil")
        h.init()
        return h

    @classmethod
    def load_hoc(cls, mod_name):
        """Loads a hoc module, available in the path.
        """
        if mod_name in cls._hocs_loaded:
            return
        h = (cls._h or cls._init())
        mod_filename = mod_name + ".hoc"
        if not h.load_file(mod_filename):
            raise RuntimeError("Cant load HOC library {}. Consider checking HOC_LIBRARY_PATH"
                               .format(mod_filename))
        cls._hocs_loaded.add(mod_name)

    @classmethod
    def require(cls, *hoc_mods):
        """Load a set of hoc mods by name.
        """
        for mod in hoc_mods:
            cls.load_hoc(mod)
        return cls._h

    @classmethod
    def load_dll(cls, dll_path):
        """Loads a Neuron mod file (typically an .so file in linux).
        """
        h = (cls._h or cls._init())
        rc = h.nrn_load_dll(dll_path)
        if rc == 0:
            raise RuntimeError("Cant load MOD dll {}. Please check LD path and dependencies"
                               .format(dll_path))

    @contextmanager
    def section_in_stack(self, sec):
        """A contect manager to push and pop a section to the Neuron stack.
        """
        sec.push()
        yield
        self.h.pop_section()

    @classmethod
    def run_sim(cls, t_stop, *monitored_sections, **params):
        """A helper to run the simulation, recording the Voltage in the specified cell sections.

        Args:
            t_stop: Stop time
            *monitored_sections: Cell sections to be probed.
            **params: Custom simulation parameters

        Returns: A simulation object
        """
        cls._h or cls._init()
        sim = Simulation(**params)
        for sec in monitored_sections:
            sim.record_activity(sec)
        sim.run(t_stop)
        return sim

    # Properties that are not found here are get / set
    # directly in neuron.h
    def __getattr__(self, item):
        # We use a cache since going down to hoc costs at least 10us
        # Cache is not expected to grow very large. Unbounded for the moment
        if item.startswith("__"):
            return object.__getattribute__(self, item)
        self._h or self._init()
        cache = self.__class__.__cache
        obj = cache.get(item)
        if obj is None:
            obj = getattr(self._h, item)
            if type(obj) is not float:
                cache[item] = obj
        return obj

    def __setattr__(self, key, value):
        try:
            object.__setattr__(self, key, value)
        except AttributeError:
            setattr(self.h, key, value)

    # public shortcuts
    HocEntity = None   # type: HocEntity
    Simulation = None  # type: Simulation
    LoadBalance = None  # type: type
    Section = None
    Segment = None


Neuron = _Neuron()
"""A singleton wrapper for the Neuron library"""


class HocEntity(object):
    _hoc_cls = None
    _hoc_obj = None
    _hoc_cldef = """
begintemplate {cls_name}
endtemplate {cls_name}
"""

    def __new__(cls, *args, **kw):
        if cls is HocEntity:
            raise TypeError("HocEntity must be subclassed")
        if cls._hoc_cls is None:
            h = Neuron.h
            # Create a HOC template to be able to use as context
            h(cls._hoc_cldef.format(cls_name=cls.__name__))
            cls._hoc_cls = getattr(h, cls.__name__)

        o = object.__new__(cls)
        o._hoc_obj = cls._hoc_cls()
        return o

    @property
    def h(self):
        return self._hoc_obj


class Simulation(object):
    # Some defaults from stdrun
    v_init = NeuronStdrunDefaults.v_init  # -65V

    def __init__(self, **args):
        args.setdefault("v_init", self.v_init)
        self.args = args
        self.t_vec = None
        self.recordings = {}

    def run(self, t_stop):
        h = Neuron.h
        self.t_vec = h.Vector()  # Time stamp vector
        self.t_vec.record(h._ref_t)

        Neuron.h.tstop = t_stop
        for key, val in self.args.items():
            setattr(Neuron.h, key, val)
        Neuron.h.run()

    def run_continue(self, t_stop):
        Neuron.h.continuerun(t_stop)

    def record_activity(self, section, rel_pos=0.5):
        if isinstance(section, Neuron.Segment):
            segment = section
            name = str(segment.sec)
        else:
            segment = section(rel_pos)
            name = section.name()

        rec_vec = Neuron.h.Vector()
        rec_vec.record(segment._ref_v)
        self.recordings[name] = rec_vec

    def get_voltages_at(self, section):
        return self.recordings[section.name()]

    def plot(self):
        try:
            from matplotlib import pyplot
        except Exception:
            logging.error("Matplotlib is not installed. Please install pyneurodamus[full]")
            return None
        if len(self.recordings) == 0:
            logging.error("No recording sections defined")
            return None
        if not self.t_vec:
            logging.error("No Simulation data. Please run it first.")
            return None

        fig = pyplot.figure()
        ax = fig.add_subplot(1, 1, 1)  # (nrows, ncols, axnum)
        for name, y in self.recordings.items():
            ax.plot(self.t_vec, y, label=name)
        ax.legend()
        fig.show()


class MComplexLoadBalancer(object):
    """Wrapper of the load balance Hoc Module with mcomplex.
    """
    def __init__(self, force_regenerate=False):
        # Can we use an existing mcomplex.dat?
        if force_regenerate or not os.path.isfile("mcomplex.dat"):
            logging.info("Generating mcomplex.dat...")
            self._create_mcomplex()
        else:
            logging.info("Using existing mcomplex.dat")
        self._lb = Neuron.h.LoadBalance()
        self._lb.read_mcomplex()

    @staticmethod
    def _create_mcomplex():
        # Save the dt of the simulation and set the dt for calculating the ExperimentalMechComplex
        # to the default value of 0.025
        prev_dt = Neuron.h.dt
        Neuron.h.dt = SimConfig.default_neuron_dt
        # ExperimentalMechComplex is a complex routine changing many state vars, and cant be reused
        # Therefore here we create a temporary LoadBalance obj
        lb = Neuron.h.LoadBalance()
        lb.ExperimentalMechComplex("StdpWA", "extracel", "HDF5", "Report", "Memory", "ASCII")
        # Revert dt to the old value
        Neuron.h.dt = prev_dt
        # mcomplex changes neuron state and results get different. We re-init
        Neuron.h.init()

    def __getattr__(self, item):
        return getattr(self._lb, item)


# shortcuts
_Neuron.HocEntity = HocEntity
_Neuron.Simulation = Simulation
_Neuron.MComplexLoadBalancer = MComplexLoadBalancer



"""
A module implementing a high-level interface to Neuron cells.
"""
from __future__ import absolute_import
import logging
from collections import defaultdict
from .configuration import GlobalConfig
from .mechanisms import Mechanism
from .synapses import _SpikeSource
from . import Neuron

__all__ = ["Cell"]


class Cell(Neuron.HocEntity, _SpikeSource):
    """
    A Cell abstraction. It allows users to instantiate Cells from morphologies or
    create them from scratch using the Cell.Builder
    """
    # We must override the basic tpl definition
    # Since the morphology parser expects several arrays
    __name__ = "Cell"
    _hoc_cldef = """
begintemplate {cls_name}
    public init, exec_within_context
    public soma, dend, apic, axon, myelin
    public synlist, all, apical, basal, somatic, axonal
    objref synlist, all, apical, basal, somatic, axonal, this
    create soma[1], dend[1], apic[1], axon[1], myelin[1]

    proc init() {{
        all     = new SectionList()
        somatic = new SectionList()
        basal   = new SectionList()
        apical  = new SectionList()
        axonal  = new SectionList()
        forall delete_section()
    }}
endtemplate {cls_name}"""

    _section_lists = ('all', 'somatic', 'axonal', 'basal', 'apical')
    _section_arrays = ('soma', 'axon', 'dend', 'apic', 'myelin')

    def __init__(self, gid=0, morpho=None):
        # type: (int, str) -> None
        self.gid = gid
        self._soma = None
        self._axon = None
        self._dend = None
        self._apic = None

        # first indexed by sec_name, then synapse_obj. Leaves are list of NetCon
        # Synapses are relative to receptor cells
        self._synapses = defaultdict(lambda: defaultdict(list))

        self._builder = None  # type: Cell.Builder.Section
        if morpho is not None:
            self.load_morphology(morpho)

    # ---
    def load_morphology(self, morpho_path, export_commands=False):
        """ Creates the cell compartments according to the given morphology
        """
        h = Neuron.require("import3d", "MorphIO")
        # try and determine format
        if morpho_path.endswith(('h5', 'H5')):
            if(export_commands):
                from neurodamus.morphio_wrapper import MorphIOWrapper
                self._commands = MorphIOWrapper(morpho_path).morph_as_hoc()
            h.morphio_read(self.h, morpho_path)
            self._soma = self.h.soma[0]

        elif morpho_path.endswith(('hoc', 'HOC')):
            h.load_file(1, morpho_path)
        else:
            if morpho_path.endswith(('asc', 'ASC')):
                imp = h.Import3d_Neurolucida3()
                if not GlobalConfig.verbosity or export_commands:
                    imp.quiet = 1
            elif morpho_path.endswith(('swc', 'SWC')):
                imp = h.Import3d_SWC_read()
            elif morpho_path.endswith(('xml', 'XML')):
                imp = h.Import3d_MorphML()
            else:
                raise ValueError(
                    "{} is not a recognised morphology file format".format(morpho_path) +
                    "Should be either .hoc, .asc, .swc, .xml!")
            try:
                imp.input(morpho_path)
                imprt = h.Import3d_GUI(imp, 0)
            except Exception:
                raise Exception("Error loading morphology. Verify Neuron outputs")

            if export_commands:
                imprt.instantiate(self.h, 1)
            else:
                imprt.instantiate(self.h)
            # Create shortcuts. Hoc arrays are fine, no need to convert
            self._soma = self.h.soma[0]
            self._dend = self.h.dend
            self._apic = self.h.apic
            self._axon = self.h.axon

            if export_commands:
                self._commands = imprt.commands

    @property
    def all(self):
        return SectionList(self.h.all)

    @property
    def soma(self):
        return self._soma

    # Other properties use neuron structures as source
    # So that we don't need to handle sync issues
    @property
    def axons(self):
        return self._axon or SectionList(self.h.axonal, self.h.axon)

    @property
    def dendrites(self):
        return self._dend or SectionList(self.h.basal, self.h.dend)

    @property
    def apical_dendrites(self):
        return self._apic or SectionList(self.h.apical, self.h.apic)

    @staticmethod
    def show_topology():
        Neuron.h.topology()

    def section_info(self, section):
        c = self.all[section] if isinstance(section, int) else section
        return ("|lenght: {} um\n".format(c.L) +
                "|diameter: {} um\n".format(c.diam) +
                "|N_segments: {}\n".format(c.nseg) +
                "|axial resistance: {} ohm.cm\n".format(c.Ra))

    # ---
    class Builder:
        """Enables building a cell from soma/axon blocks
        """
        class Section:
            SOMA = 0
            DENDRITE = 1
            APICAL_DENDRITE = 2
            AXON = 3

            def __init__(self, name, length, n_segments=None, sec_type=None, **params):
                # type: (str, float, int, int, dict) -> None
                """Creates a new section
                Args:
                    name: Section name
                    **params: Additional properties to be set on the hoc object
                """
                self.parent = None
                self.sec_type = sec_type
                self.this = Neuron.h.Section(name=name)
                self.this.L = length
                if n_segments:
                    self.this.nseg = n_segments
                for param, value in params.items():
                    setattr(self.this, param, value)
                self.sub_nodes = []

            def add(self, name, length, n_segments, **params):
                """Creates a new section as a child of the current section"""
                self.attach(self.__class__(name, length, nseg=n_segments, **params))
                return self

            def add_dendrite(self, name, length, n_segments, apical=False, **params):
                """Creates the first section of a dendrite"""
                self._ensure_soma()
                t = self.APICAL_DENDRITE if apical else self.DENDRITE
                return self.add(name, length, n_segments, sec_type=t, **params)

            def add_axon(self, name, length, n_segments, **params):
                """Creates the first section of an axon"""
                self._ensure_soma()
                return self.add(name, length, n_segments, sec_type=self.AXON, **params)

            def attach(self, *nodes):
                """Adds the given sections as children of the current"""
                for n in nodes:
                    self.sub_nodes.append(n)
                    n.set_parent(self)
                return self

            def append(self, name, length, n_segments, **params):
                """Creates a new section as a child and returns it"""
                newnode = self.__class__(name, length, nseg=n_segments, **params)
                self.attach(newnode)
                return newnode

            def append_dendrite(self, name, length, n_segments, apical=False, **params):
                """Creates the first section of an axon"""
                self._ensure_soma()
                t = self.APICAL_DENDRITE if apical else self.DENDRITE
                return self.append(name, length, n_segments, sec_type=t, **params)

            def append_axon(self, name, length, n_segments, **params):
                """Creates the first section of an axon"""
                self._ensure_soma()
                return self.append(name, length, n_segments, sec_type=self.AXON, **params)

            def chain(self, *nodes):
                """Chain given nodes in parent-child relations, and make it child of the current"""
                parent = self
                for n in nodes:
                    n.set_parent(parent)
                    parent = n
                return parent

            def set_parent(self, parent):
                """Sets the parent of the given section"""
                self.parent = parent
                self.this.connect(parent.this)
                return self

            def get_root(self):
                """Finds the root section by crawling parents up"""
                sec = self
                while isinstance(sec.parent, self.__class__):
                    sec = sec.parent
                return sec

            def create(self):
                """Builds the cell the current section belongs to.
                If no root is found (e.g. disconnected branch) an exception is raised
                """
                sec = self.get_root()  # type: self.__class__
                if sec.parent is None:
                    raise CellCreationError("Disconnected subtree. Attach to a CellBuilder root "
                                            "node")
                # If parent is True we must create the cell. Otherwise use it
                c = Cell() if sec.parent is True else sec.parent
                c.h.all.wholetree(sec=sec.this)

                for branch_sec in sec.sub_nodes:
                    node = branch_sec.this
                    t = branch_sec.sec_type
                    if t == self.SOMA:
                        c.h.somatic.subtree(sec=node)
                        print("Somas: I got", list(c.h.somatic))
                    elif t == self.DENDRITE:
                        c.h.basal.subtree(sec=node)
                        c._dend = list(c.h.basal)
                    elif t == self.APICAL_DENDRITE:
                        c.h.apical.subtree(sec=node)
                        c._apic = list(c.h.apical)
                    elif t == self.AXON:
                        c.h.axonal.subtree(sec=node)
                        c._axon = list(c.h.axonal)
                    else:
                        logging.warning("Branch starting at %s doesnt have a type", node)
                c._soma = sec.this
                c._builder = sec
                return c

            def _ensure_soma(self):
                if self.sec_type is not self.SOMA:
                    raise CellCreationError("Dendrites must start on the soma")

        class DendriteSection(Section):
            def __init__(self, name, length, n_segments=None, apical=False, **params):
                Cell.Builder.Section.__init__(
                    self, name, length, n_segments,
                    sec_type=self.APICAL_DENDRITE if apical else self.DENDRITE,
                    **params)

        class AxonSection(Section):
            def __init__(self, name, length, n_segments=None, **params):
                Cell.Builder.Section.__init__(self, name, length, n_segments, **params)

        @classmethod
        def add_soma(cls, diam, name="soma", **params):
            root = cls.Section(name, length=diam, diam=diam, sec_type=cls.Section.SOMA, **params)
            root.parent = True  # this is root
            return root

    # ---
    def init_soma(self, diam, name="soma", **params):
        """Creates a soma and returns the section builder
        NOTE: you must call create at the end so that the new sections are added to the cell
        """
        soma = self.Builder.add_soma(diam, name, **params)
        soma.parent = self
        return soma

    @property
    def builder(self):
        """Returns the Section builder object, to build additional axon/dendrites"""
        if self._builder is None:
            raise RuntimeError("When the cell is void start creating it with init_soma()")
        return self._builder

    # --------
    # Synapses
    # --------

    def add_synapse(self, src_seg, target_seg, syn_props_obj, **conn_options):
        # type: (Neuron.nrn.Segment, Neuron.nrn.Segment, object, **dict) -> object
        """Adds an incoming synapse from another cell, according to the options."""
        synapse = self.add_synaptic_receptor(syn_props_obj, target_seg)
        netcon = self._add_connection(src_seg, synapse, **conn_options)
        self._synapses[target_seg.sec.name()][synapse].append(netcon)
        return netcon

    def add_synaptic_receptor(self, target_seg, syn_props_obj, **custom):
        """Creates a raw neuron Synapse"""
        synapse = getattr(Neuron.h, type(syn_props_obj).__name__)(target_seg)
        # Applies specific properties
        syn_props_obj.apply(synapse, **custom)
        # Save in the cell synapse dict
        self._synapses[target_seg.sec.name()][synapse] = []
        return synapse

    def connect_to(self, synapse_receptor, threshold=None, delay=None, weight=None):
        """Creates a synapse between the current cell soma extremity and a given synapse receptor
        Notes: This is a relatively low-level f, there is no automatic registration of the netcon
        """
        return self._add_connection(self.soma(1), synapse_receptor,
                                    threshold=threshold, delay=delay, weight=weight)

    @staticmethod
    def _add_connection(src_segment, synapse_receptor, **props):
        nc = Neuron.h.NetCon(src_segment._ref_v, synapse_receptor, sec=src_segment.sec)
        return _SpikeSource._setup_netcon(nc, **props)

    # Declare shortcut
    Mechanisms = Mechanism


class SectionList(object):
    """A SectionList wrapper providing convenience methods, inc len(),
       and consolidating SectionList and hoc section arrays
    """
    __slots__ = ('_hlist', '_harray')

    def __init__(self, hoc_section_list, hoc_section_array=None):
        self._hlist = hoc_section_list
        self._harray = hoc_section_array

    def __len__(self):
        return sum(1 for _ in self._hlist)

    def __getattr__(self, item):
        return object.__getattribute__(self._hlist, item)

    def __getitem__(self, item):
        if self._harray is not None:
            return self._harray[item]
        else:
            logging.info("Indexing list without array. This might be inneficient")
            for i, elem in enumerate(self._hlist):
                if i == item:
                    return elem
            return None

    def __iter__(self):
        return iter(self._hlist)


class CellCreationError(Exception):
    """ An exception for errors in instantiating a cell
    """



"""
Compatibility Classes to work similar to HOC types, recreating or wrapping them
"""
from __future__ import absolute_import
from array import array
try:
    import collections.abc as collections_abc  # Py >= 3.3
except ImportError:
    import collections as collections_abc


class Vector(array):
    """Behavior similar to Hoc Vector
    """
    __slots__ = ()

    def __new__(cls, type_="I", array=[]):
        return super().__new__(Vector, type_, array)

    def size(self):
        return len(self)

    @property
    def x(self):
        return self

    def __add__(self, other):
        array.extend(self, other)
        return self

    def as_hoc(self):
        """When API compat is not enough, convert to a true hov Vector"""
        return hoc_vector(self)


class List(list):
    """Behavior similar to Hoc List
    """
    __slots__ = ()

    def count(self, obj=None):
        return super().count(obj) if obj else len(self)

    def o(self, idx):
        return self[int(idx)]


class Map(collections_abc.Mapping):
    """Class which bring Python map API to hoc Maps
    """
    __slots__ = ('_hoc_map', '_size', 'String')

    def __new__(cls, wrapped_obj, *args, **kwargs):
        """ If the wrapped entity is not an hoc map, but a Python dict
            then also wrap it using PyMap for a similar API
        """
        if isinstance(wrapped_obj, dict):
            return PyMap(wrapped_obj)
        return object.__new__(cls)

    def __init__(self, hoc_map):
        self._hoc_map = hoc_map
        self._size = int(hoc_map.count())
        from neuron import h
        self.String = h.String

    def __iter__(self):
        return (self._hoc_map.key(i).s for i in range(self._size))

    def items(self):
        return ((self._hoc_map.key(i).s, self._hoc_map.o(i))
                for i in range(self._size))

    def values(self):
        return (self._hoc_map.o(i) for i in range(self._size))

    keys = __iter__

    def get(self, key, default=None):
        return self[key] if key in self else default

    def __getitem__(self, item):
        value = self._hoc_map.get(item)
        if hasattr(value, 's'):  # hoc strings have the value in .s attribute
            value = value.s
        return value

    def __setitem__(self, key, value):
        if self._hoc_map.exists(key):
            self._hoc_map.get(key).s = str(value)
        else:
            self._hoc_map.put(key, self.String(str(value)))
            self._size = int(self._hoc_map.count())  # update size

    def update(self, other_map):
        for key, val in other_map.items():
            self[key] = val

    def __contains__(self, item):
        return self._hoc_map.exists(item) > 0

    def __len__(self):
        return self._size

    @property
    def hoc_map(self):
        """Returns the raw hoc map
        """
        return self._hoc_map

    def as_dict(self, parse_strings=False):
        """Creates a real dictionary from the Map.

        Args:
            parse_strings: If true converts string objects in both key and values to
                real strings (xx.s) and attempts to convert values to float
        """
        if parse_strings:
            def parse(stri):
                try:
                    return float(stri)
                except ValueError:
                    return stri
            new_map = {key: parse(val.s) for key, val in self.items()}
        else:
            new_map = dict(self)
        new_map["_hoc"] = self.hoc_map
        return new_map


class PyMap(dict):
    """
    PyMap does basically the reverse of compat.Map: it's a true dict but capable of
    getting a hoc map, built on the fly
    """
    __slots__ = ()

    @property
    def hoc_map(self):
        """Returns the raw hoc map
        """
        return self._dict_as_hoc(self)

    @classmethod
    def _value_as_hoc(cls, value):
        from neuron import h
        if isinstance(value, dict):
            return cls._dict_as_hoc(value)
        return h.String(str(value))

    @classmethod
    def _dict_as_hoc(cls, d):
        from neuron import h
        m = h.Map()
        for key, val in d.items():
            m.put(h.String(key), cls._value_as_hoc(val))
        return m

    def as_dict(self, *_, **_kw):
        return self


def hoc_vector(np_array):
    from neuron import h
    hoc_vec = h.Vector(np_array.size)
    hoc_vec.as_numpy()[:] = np_array
    return hoc_vec



import os
import logging
from pathlib import Path
from ._utils import run_only_rank0
from . import NeurodamusCore as Nd


class _CoreNEURONConfig(object):
    """
    The CoreConfig class is responsible for managing the configuration of the CoreNEURON simulation.
    It writes the simulation / report configurations and calls the CoreNEURON solver.
    """
    sim_config_file = "sim.conf"
    report_config_file = "report.conf"
    output_root = "output"
    datadir = f"{output_root}/coreneuron_input"
    default_cell_permute = 0
    artificial_cell_object = None

    # Instantiates the artificial cell object for CoreNEURON
    # This needs to happen only when CoreNEURON simulation is enabled
    def instantiate_artificial_cell(self):
        self.artificial_cell_object = Nd.CoreNEURONArtificialCell()

    @run_only_rank0
    def write_report_config(
            self, report_name, target_name, report_type, report_variable,
            unit, report_format, target_type, dt, start_time, end_time, gids,
            buffer_size=8):
        import struct
        num_gids = len(gids)
        logging.info(f"Adding report {report_name} for CoreNEURON with {num_gids} gids")
        report_conf = Path(self.output_root) / self.report_config_file
        report_conf.parent.mkdir(parents=True, exist_ok=True)
        with report_conf.open("ab") as fp:
            # Write the formatted string to the file
            fp.write(("%s %s %s %s %s %s %d %lf %lf %lf %d %d\n" % (
                report_name,
                target_name,
                report_type,
                report_variable,
                unit,
                report_format,
                target_type,
                dt,
                start_time,
                end_time,
                num_gids,
                buffer_size
            )).encode())
            # Write the array of integers to the file in binary format
            fp.write(struct.pack(f'{num_gids}i', *gids))
            fp.write(b'\n')

    @run_only_rank0
    def write_sim_config(
            self, tstop, dt, forwardskip, prcellgid, celsius, v_init,
            pattern=None, seed=None, model_stats=False, enable_reports=True):
        simconf = Path(self.output_root) / self.sim_config_file
        logging.info(f"Writing sim config file: {simconf}")
        simconf.parent.mkdir(parents=True, exist_ok=True)
        with simconf.open("w") as fp:
            fp.write(f"outpath='{os.path.abspath(self.output_root)}'\n")
            fp.write(f"datpath='{os.path.abspath(self.datadir)}'\n")
            fp.write(f"tstop={tstop}\n")
            fp.write(f"dt={dt}\n")
            fp.write(f"forwardskip={forwardskip}\n")
            fp.write(f"prcellgid={int(prcellgid)}\n")
            fp.write(f"celsius={celsius}\n")
            fp.write(f"voltage={v_init}\n")
            fp.write(f"cell-permute={int(self.default_cell_permute)}\n")
            if pattern:
                fp.write(f"pattern='{pattern}'\n")
            if seed:
                fp.write(f"seed={int(seed)}\n")
            if model_stats:
                fp.write("'model-stats'\n")
            if enable_reports:
                fp.write(f"report-conf='{self.output_root}/{self.report_config_file}'\n")
            fp.write("mpi=true\n")

    @run_only_rank0
    def write_report_count(self, count, mode="w"):
        report_config = Path(self.output_root) / self.report_config_file
        report_config.parent.mkdir(parents=True, exist_ok=True)
        with report_config.open(mode) as fp:
            fp.write(f"{count}\n")

    @run_only_rank0
    def write_population_count(self, count):
        self.write_report_count(count, mode="a")

    @run_only_rank0
    def write_spike_population(self, population_name, population_offset=None):
        report_config = Path(self.output_root) / self.report_config_file
        report_config.parent.mkdir(parents=True, exist_ok=True)
        with report_config.open("a") as fp:
            fp.write(population_name)
            if population_offset is not None:
                fp.write(f" {int(population_offset)}")
            fp.write("\n")

    @run_only_rank0
    def write_spike_filename(self, filename):
        report_config = Path(self.output_root) / self.report_config_file
        report_config.parent.mkdir(parents=True, exist_ok=True)
        with report_config.open("a") as fp:
            fp.write(filename)
            fp.write("\n")

    def psolve_core(self, save_path=None, restore_path=None):
        from neuron import coreneuron
        from . import NeurodamusCore as Nd

        Nd.cvode.cache_efficient(1)
        coreneuron.enable = True
        coreneuron.file_mode = True
        coreneuron.sim_config = f"{self.output_root}/{self.sim_config_file}"
        if save_path:
            coreneuron.save_path = save_path
        if restore_path:
            coreneuron.restore_path = restore_path
        # Model is already written to disk by calling pc.nrncore_write()
        coreneuron.skip_write_model_to_disk = True
        coreneuron.model_path = f"{self.datadir}"
        Nd.pc.psolve(Nd.tstop)


# Singleton
CoreConfig = _CoreNEURONConfig()



"""
High-Level wrapper to Neuron's cell synapse mechanisms
"""
from ..utils import ConfigT
from . import Neuron


# -------------------------
# Synapse (receptor) config
# -------------------------

class _SynapseReceptor(ConfigT):
    """A synapse configuration. Can then be instantiated among several pre-post pairs
    """
    e = None    # mV
    i = None    # nA

    def __new__(cls, segment=None, **properties):
        if cls is _SynapseReceptor:
            raise TypeError("Can't instantiate an abstract Synapse. Use a specific type instead")
        if segment is not None:
            segment = segment(0.5) if isinstance(segment, Neuron.Section) else segment
            syn = getattr(Neuron.h, cls.__name__)(segment)
            [setattr(syn, key, val) for key, val in properties.items()]
            return syn
        return object.__new__(cls)

    def create_on(self, *segments):
        return [self.__class__(seg, **self.as_dict()) for seg in segments]


# Specific Synapse Types
# NOTE: The name of the specific Synapse Makers must match that of the Hoc class

class AlphaSynapse(_SynapseReceptor):
    onset = None    # ms
    tau  = None     # ms
    gmax = None     # umho


class ExpSyn(_SynapseReceptor):
    tau = None  # ms


class Exp2Syn(_SynapseReceptor):
    tau1 = None  # ms
    tau2 = None  # ms


# ----------------------------------
# Synapse sources (Spike generators)
# ----------------------------------

class _SpikeSource(object):
    def connect_to(self, synapse_receptor, weights=None, threshold=None, delay=None):
        """Creates a synapse connection"""
        raise NotImplementedError()

    @staticmethod
    def _setup_netcon(netcon, weight=None, **props):
        for key, val in props.items():
            if val is not None and key in ["threshold", "delay"]:
                setattr(netcon, key, val)
        if isinstance(weight, dict):
            for idx, val in weight.items():
                netcon.weight[idx] = val
        else:
            netcon.weight[0] = weight
        return netcon


class VirtualSpikeSource(ConfigT, _SpikeSource):
    """
    Uses Neuron NetStim to create an artificial spike source
    """
    interval = None     # ms (mean) time between spikes
    number = None       # number of spikes
    start = None        # ms (most likely) start time of first spike
    noise = None        # range 0 to 1. Fractional randomness. (negexp distribution)

    def __init__(self, interval=None, number=None, start=None, noise=None, manage_objs=True):
        """ Creates an artificial spike generator
        Args:
            manage_objs: NetCon and Synapse receptors are stored internally to avoid
                garbage-collection at the expense of some extra memory [default: True]
        """
        self._netstim = Neuron.h.NetStim()
        ConfigT.__init__(self, **{"interval": interval, "number": number,
                                  "start": start, "noise": noise})
        self.apply(self._netstim)
        self._managed_objs = manage_objs
        self._objs = []

    def connect_to(self, synapse_receptor, weight=None, threshold=None, delay=None):
        netcon = Neuron.h.NetCon(self._netstim, synapse_receptor)
        self._setup_netcon(netcon, weight, threshold=threshold, delay=delay)
        self._objs.append((netcon, synapse_receptor))
        return netcon


# Alias
NetStim = VirtualSpikeSource


# -------------------------------------
# Full config of a synapse
# -------------------------------------
class CellSynapse(_SpikeSource, _SynapseReceptor):
    _netcon_fields = ("threshold", "delay", "weight")

    def get_synpase_conf(self):
        self.as_dict(excludes=self._netcon_fields)

    def get_netcon_conf(self):
        self.as_dict(subset=self._netcon_fields)



"""
Runtime configuration
"""
from __future__ import absolute_import
import logging
import os
import os.path
import re
from collections import defaultdict
from enum import Enum

from ..io.config_parser import BlueConfig
from ..io.sonata_config import SonataConfig
from ..utils import compat
from ..utils.logging import log_verbose
from ..utils.pyutils import ConfigT
from ._shmutils import SHMUtil

EXCEPTION_NODE_FILENAME = ".exception_node"
"""A file which controls which rank shows exception"""


class LogLevel:
    ERROR_ONLY = 0
    DEFAULT = 1
    VERBOSE = 2
    DEBUG = 3


class ConfigurationError(Exception):
    """
    Error due to invalid settings in BlueConfig
    ConfigurationError should be raised by all ranks to be caught
    properly. Otherwise we might end up with deadlocks.
    For Exceptions that are raised by a single rank Exception
    should be used.
    This is due to the way Exceptions are caught from commands.py.
    """
    pass


class GlobalConfig:
    verbosity = LogLevel.DEFAULT
    debug_conn = os.getenv('ND_DEBUG_CONN', [])
    if debug_conn:
        debug_conn = [int(gid) for gid in os.getenv('ND_DEBUG_CONN', '').split(',')]
        verbosity = 3

    @classmethod
    def set_mpi(cls):
        os.environ["NEURON_INIT_MPI"] = "1"


class Feature(Enum):
    Replay = 1
    SpontMinis = 2
    SynConfigure = 3
    Stimulus = 4
    LoadBalance = 5


class CliOptions(ConfigT):
    build_model = None
    simulate_model = True
    model_path = None  # Currently is output-path
    output_path = None
    keep_build = False
    lb_mode = None
    modelbuilding_steps = None
    experimental_stims = False
    enable_coord_mapping = False
    save = False
    save_time = None
    restore = None
    enable_shm = False
    model_stats = False
    simulator = None
    dry_run = False

    # Restricted Functionality support, mostly for testing

    class NoRestriction:
        """Provide container API, where `in` checks are always True"""
        def __contains__(self, _) -> bool:
            return True

    NoRestriction = NoRestriction()  # Singleton

    restrict_features = NoRestriction  # can also be a list
    restrict_node_populations = NoRestriction
    restrict_connectivity = 0          # no restriction, 1 to disable projections, 2 to disable all
    restrict_stimulus = NoRestriction  # possible list of Stim names


class CircuitConfig(ConfigT):
    _name = None
    Engine = None
    CircuitPath = ConfigT.REQUIRED
    nrnPath = ConfigT.REQUIRED
    CellLibraryFile = None
    MEComboInfoFile = None
    METypePath = None
    MorphologyType = None
    MorphologyPath = None
    CircuitTarget = None
    PopulationID = 0
    DetailedAxon = False


class RNGConfig(ConfigT):
    Modes = Enum("Mode", "COMPATIBILITY RANDOM123 UPMCELLRAN4")
    mode = Modes.COMPATIBILITY
    global_seed = None
    IonChannelSeed = None
    StimulusSeed = None
    MinisSeed = None
    SynapseSeed = None


class NeuronStdrunDefaults:
    """Neuron stdrun default (src: share/lib/hoc/stdrun.hoc"""
    using_cvode_ = 0
    stdrun_quiet = 0
    realtime = 0
    tstop = 5
    stoprun = 0
    steps_per_ms = 1 / .025
    nstep_steprun = 1
    global_ra = 35.4
    v_init = -65


class LoadBalanceMode(Enum):
    """An enumeration, inc parser, of the load balance modes.
    """
    RoundRobin = 0
    WholeCell = 1
    MultiSplit = 2

    @classmethod
    def parse(cls, lb_mode):
        """Parses the load balancing mode from a string.
        Options other than WholeCell or MultiSplit are considered RR
        """
        if lb_mode is None:
            return None
        _modes = {
            "rr": cls.RoundRobin,
            "roundrobin": cls.RoundRobin,
            "wholecell": cls.WholeCell,
            "loadbalance": cls.MultiSplit,
            "multisplit": cls.MultiSplit
        }
        lb_mode_enum = _modes.get(lb_mode.lower())
        if lb_mode_enum is None:
            raise ConfigurationError("Unknown load balance mode: " + lb_mode)
        return lb_mode_enum

    class AutoBalanceModeParams:
        """Parameters for auto-selecting a load-balance mode"""
        multisplit_cpu_cell_ratio = 4  # For warning
        cell_count = 1000
        duration = 1000
        mpi_ranks = 200

    @classmethod
    def auto_select(cls, use_neuron, cell_count, duration, auto_params=AutoBalanceModeParams):
        """Simple heuristics for auto selecting load balance"""
        from ._neurodamus import MPI
        lb_mode = LoadBalanceMode.RoundRobin
        reason = ""
        if MPI.size == 1:
            lb_mode = LoadBalanceMode.RoundRobin
            reason = "Single rank - not worth using Load Balance"
        elif use_neuron and MPI.size >= auto_params.multisplit_cpu_cell_ratio * cell_count:
            logging.warn("There's potentially a high number of empty ranks. "
                         "To activate multi-split set --lb-mode=MultiSplit")
        elif (cell_count > auto_params.cell_count
              and duration > auto_params.duration
              and MPI.size > auto_params.mpi_ranks):
            lb_mode = LoadBalanceMode.WholeCell
            reason = 'Simulation size'
        return lb_mode, reason


class _SimConfig(object):
    """
    A class initializing several HOC config objects and proxying to simConfig
    """
    __slots__ = ()
    config_file = None
    cli_options = None
    run_conf = None
    output_root = None
    base_circuit = None
    extra_circuits = None
    sonata_circuits = None
    projections = None
    connections = None
    stimuli = None
    injects = None
    reports = None
    configures = None
    modifications = None

    # Hoc objects used
    _config_parser = None
    _parsed_run = None
    _blueconfig = None  # new python BlueConfig parser
    _simconf = None
    rng_info = None

    # In principle not all vars need to be required as they'r set by the parameter functions
    blueconfig_dir = None
    current_dir = None
    default_neuron_dt = 0.025
    buffer_time = 25
    save = None
    save_time = None
    restore = None
    extracellular_calcium = None
    secondorder = None
    use_coreneuron = False
    use_neuron = True
    coreneuron_datadir = None
    coreneuron_outputdir = None
    corenrn_buff_size = 8
    delete_corenrn_data = False
    modelbuilding_steps = 1
    build_model = True
    simulate_model = True
    loadbal_mode = None
    synapse_options = {}
    is_sonata_config = False
    spike_location = "soma"
    spike_threshold = -30
    dry_run = False

    _validators = []
    _requisitors = []
    _cell_requirements = {}

    restore_coreneuron = property(lambda self: self.use_coreneuron and bool(self.restore))
    cell_requirements = property(lambda self: self._cell_requirements)

    @classmethod
    def init(cls, config_file, cli_options):
        # Import these objects scope-level to avoid cross module dependency
        from . import NeurodamusCore as Nd
        Nd.init()
        if not os.path.isfile(config_file):
            raise ConfigurationError("Config file not found: " + config_file)
        logging.info("Initializing Simulation Configuration and Validation")
        cls.is_sonata_config = config_file.endswith(".json")

        log_verbose("ConfigFile: %s", config_file)
        log_verbose("CLI Options: %s", cli_options)
        cls.config_file = config_file
        cls._config_parser = cls._init_config_parser(config_file)
        cls._parsed_run = compat.Map(cls._config_parser.parsedRun)  # easy access to hoc Map
        if not cls.is_sonata_config:
            cls._blueconfig = BlueConfig(config_file)
        else:
            cls._blueconfig = cls._config_parser   # Please refactor me
            cls.sonata_circuits = cls._config_parser.circuits
        cls.blueconfig_dir = os.path.dirname(os.path.abspath(config_file))

        cls.projections = compat.Map(cls._config_parser.parsedProjections)
        cls.connections = compat.Map(cls._config_parser.parsedConnects)
        cls.stimuli = compat.Map(cls._config_parser.parsedStimuli)
        cls.injects = compat.Map(cls._config_parser.parsedInjects)
        cls.reports = compat.Map(cls._config_parser.parsedReports)
        cls.configures = compat.Map(cls._config_parser.parsedConfigures or {})
        cls.modifications = compat.Map(cls._config_parser.parsedModifications or {})
        cls.cli_options = CliOptions(**(cli_options or {}))
        cls.dry_run = cls.cli_options.dry_run
        # change simulator by request before validator and init hoc config
        if cls.cli_options.simulator:
            cls._parsed_run["Simulator"] = cls.cli_options.simulator

        cls.run_conf = run_conf = cls._parsed_run.as_dict(parse_strings=True)
        for validator in cls._validators:
            validator(cls, run_conf)

        logging.info("Checking simulation requirements")
        for requisitor in cls._requisitors:
            requisitor(cls, cls._config_parser)

        logging.info("Initializing hoc config objects")
        if not cls.is_sonata_config:
            cls._parsed_run.update(run_conf)  # sync hoc config
        cls._init_hoc_config_objs()

    @classmethod
    def get_blueconfig_hoc_section(cls, section_name):
        # Sonata config sections are stored as PyMap, so we need to convert first
        if isinstance(cls._config_parser, SonataConfig):
            section = getattr(cls._config_parser, section_name)
            return section and section.hoc_map
        return getattr(cls._config_parser, section_name)

    @classmethod
    def _init_config_parser(cls, config_file):
        if not config_file.endswith(".json"):
            config_parser = cls._init_hoc_config_parser(config_file)  # legacy reader
        else:
            config_parser = SonataConfig(config_file)
        if config_parser.parsedRun is None:
            raise ConfigurationError("No 'Run' block found in BlueConfig %s", config_file)
        return config_parser

    @classmethod
    def _init_hoc_config_parser(cls, config_file):
        from . import NeurodamusCore as Nd
        config_parser = Nd.ConfigParser()
        config_parser.open(config_file)
        if Nd.pc.id == 0:
            config_parser.toggleVerbose()
        return config_parser

    @classmethod
    def _init_hoc_config_objs(cls):
        """Init objects which parse/check configs in the hoc world"""
        from neuron import h
        parsed_run = cls._parsed_run.hoc_map
        cls._simconf = h.simConfig
        cls._simconf.interpret(parsed_run)

        cls.rng_info = h.RNGSettings()
        cls.rng_info.interpret(parsed_run)
        if parsed_run.exists("BaseSeed"):
            logging.info("User-defined RNG base seed %s", parsed_run.valueOf("BaseSeed"))

    @classmethod
    def validator(cls, f):
        """Decorator to register parameters / config validators"""
        cls._validators.append(f)

    @classmethod
    def requisitor(cls, f):
        """Decorator to register requirements investigators"""
        cls._requisitors.append(f)

    @classmethod
    def update_connection_blocks(cls, alias):
        """Convert source destination to real population names

        Args:
            alias: A dict associating alias->real_name's
        """
        from ..target_manager import TargetSpec  # avoid cyclic deps

        restrict_features = SimConfig.cli_options.restrict_features
        if Feature.SpontMinis not in restrict_features:
            logging.warning("Disabling SpontMinis (restrict_features)")
        if Feature.SynConfigure not in restrict_features:
            logging.warning("Disabling SynConfigure (restrict_features)")

        def update_item(conn, item):
            src_spec = TargetSpec(conn.get(item))
            src_spec.population = alias.get(src_spec.population, src_spec.population)
            conn[item] = str(src_spec)

        new_connections = {}
        for name, conn in cls.connections.items():
            conn = compat.Map(conn).as_dict(True)
            update_item(conn, "Source")
            update_item(conn, "Destination")
            if Feature.SpontMinis not in restrict_features:
                conn.pop("SpontMinis", None)
            if Feature.SynConfigure not in restrict_features:
                conn.pop("SynapseConfigure", None)

            new_connections[name] = conn

        cls.connections = new_connections

    @classmethod
    def check_connections_configure(cls, target_manager):
        check_connections_configure(cls, target_manager)  # top_level

    @classmethod
    def get_stim_inject(cls, stim_name):
        for _, inject in cls.injects.items():
            inject = compat.Map(inject).as_dict(parse_strings=True)
            if stim_name == inject["Stimulus"]:
                return inject
        return None

    @classmethod
    def check_cell_requirements(cls, target_manager):
        _input_resistance(cls, target_manager)  # top_level


# Singleton
SimConfig = _SimConfig()


def find_input_file(filepath, search_paths=(), alt_filename=None):
    """Determine the full path of input files.

    Relative paths are built from Run configuration entries, and never pwd.
    In case filepath points to a file, alt_filename is disregarded

    Args:
        filepath: The relative or absolute path of the file to find
        path_conf_entries: (tuple) Run configuration entries to build the absolute path
        alt_filename: When the filepath is a directory, attempt finding a given filename
    Returns:
        The absolute path to the data file
    Raises:
        (ConfigurationError) If the file could not be found
    """
    search_paths += (SimConfig.current_dir, SimConfig.blueconfig_dir)

    def try_find_in(fullpath):
        if os.path.isfile(fullpath):
            return fullpath
        if not os.path.exists(fullpath):
            return None
        if alt_filename is not None:
            alt_file_path = os.path.join(fullpath, alt_filename)
            if os.path.isfile(alt_file_path):
                return alt_file_path
        logging.warning("Deprecated: Data source found is not a file")
        return fullpath

    if os.path.isabs(filepath):
        # if it's absolute path then can be used immediately
        file_found = try_find_in(filepath)
    else:
        file_found = None
        for path in search_paths:
            file_found = try_find_in(os.path.join(path, filepath))
            if file_found:
                break

    if not file_found:
        raise ConfigurationError("Could not find file %s" % filepath)

    logging.debug("data file %s path: %s", filepath, file_found)
    return file_found


def _check_params(section_name, data, required_fields,
                  numeric_fields=(), non_negatives=(),
                  valid_values=None, deprecated_values=None):
    """Generic function to check a dict-like data set conforms to the field prescription
    """
    for param in required_fields:
        if param not in data:
            raise ConfigurationError("BlueConfig mandatory param not present: [%s] %s"
                                     % (section_name, param))
    for param in set(numeric_fields + non_negatives):
        val = data.get(param)
        try:
            val and float(val)
        except ValueError:
            raise ConfigurationError("BlueConfig param must be numeric: [%s] %s"
                                     % (section_name, param))
    for param in non_negatives:
        val = data.get(param)
        if val and float(val) < 0:
            raise ConfigurationError("BlueConfig param must be positive: [%s] %s"
                                     % (section_name, param))

    for param, valid in (valid_values or {}).items():
        val = data.get(param)
        if val and val not in valid:
            raise ConfigurationError("BlueConfig param value is invalid: [%s] %s = %s"
                                     % (section_name, param, val))

    for param, deprecated in (deprecated_values or {}).items():
        val = data.get(param)
        if val and val in deprecated:
            logging.warning("BlueConfig param value is deprecated: [%s] %s = %s"
                            % (section_name, param, val))


@SimConfig.validator
def _run_params(config: _SimConfig, run_conf):
    required_fields = ("Duration",)
    numeric_fields = ("BaseSeed", "StimulusSeed", "Celsius", "V_Init")
    non_negatives = ("Duration", "Dt", "ModelBuildingSteps", "ForwardSkip")
    _check_params("Run default", run_conf, required_fields, numeric_fields, non_negatives)


@SimConfig.validator
def _loadbal_mode(config: _SimConfig, run_conf):
    cli_args = config.cli_options
    if Feature.LoadBalance not in cli_args.restrict_features:
        logging.warning("Disabled Load Balance (restrict_features)")
        config.loadbal_mode = LoadBalanceMode.RoundRobin
        return
    lb_mode_str = cli_args.lb_mode or run_conf.get("RunMode")
    config.loadbal_mode = LoadBalanceMode.parse(lb_mode_str)


@SimConfig.validator
def _projection_params(config: _SimConfig, run_conf):
    required_fields = ("Path",)
    non_negatives = ("PopulationID",)
    for name, proj in config.projections.items():
        _check_params("Projection " + name, compat.Map(proj), required_fields, (), non_negatives)
        _validate_file_extension(compat.Map(proj).get("Path"))


@SimConfig.validator
def _stimulus_params(config: _SimConfig, run_conf):
    required_fields = ("Mode", "Pattern", "Duration", "Delay",)
    numeric_fields = ("Dt", "RiseTime", "DecayTime", "AmpMean", "AmpVar", "MeanPercent",
                      "SDPercent", "AmpCV", "AmpStart", "AmpEnd", "PercentStart",
                      "PercentEnd", "PercentLess", "Mean", "Variance", "Voltage", "RS",)
    non_negatives = ("Duration", "Delay", "Rate", "Frequency", "Width", "Lambda", "Weight",
                     "NumOfSynapses", "Seed",)
    valid_values = {
        "Mode": ("Current", "Voltage", "Conductance", "spikes"),
        "Pattern": {
            "Hyperpolarizing", "Linear", "Noise", "Pulse", "RelativeLinear",
            "RelativeShotNoise", "SEClamp", "ShotNoise", "Sinusoidal",
            "SubThreshold", "SynapseReplay", "OrnsteinUhlenbeck",
            "NPoisson", "NPoissonInhomogeneous", "ReplayVoltageTrace",
            "AbsoluteShotNoise", "RelativeOrnsteinUhlenbeck"
        }
    }
    deprecated_values = {
        "Pattern": ("NPoisson", "NPoissonInhomogeneous", "ReplayVoltageTrace")
    }
    for name, stim in config.stimuli.items():
        _check_params("Stimulus " + name, compat.Map(stim), required_fields, numeric_fields,
                      non_negatives, valid_values, deprecated_values)


@SimConfig.validator
def _modification_params(config: _SimConfig, run_conf):
    required_fields = ("Target", "Type",)
    for name, mod_block in config.modifications.items():
        _check_params("Modification " + name, compat.Map(mod_block), required_fields, ())


def _make_circuit_config(config_dict, req_morphology=True):
    if config_dict.get("CircuitPath") == "<NONE>":
        config_dict["CircuitPath"] = False
        config_dict["nrnPath"] = False
        config_dict["MorphologyPath"] = False
    elif config_dict.get("nrnPath") == "<NONE>":
        config_dict["nrnPath"] = False
        if config_dict.get("CellLibraryFile", "start.ncs") == "start.ncs":
            raise ConfigurationError(
                "ncs circuits don't support disabling connectivity with nrnPath='<NONE>'")
    _validate_circuit_morphology(config_dict, req_morphology)
    _validate_file_extension(config_dict.get("CellLibraryFile"))
    _validate_file_extension(config_dict.get("nrnPath"))
    return CircuitConfig(config_dict)


def _validate_circuit_morphology(config_dict, required=True):
    morph_path = config_dict.get("MorphologyPath")
    morph_type = config_dict.get("MorphologyType")
    if morph_path is None and required:
        raise ConfigurationError("No morphology path provided (Required!)")
    # Some circuit types may not require morphology files
    if not morph_path:
        log_verbose(" > Morphology src: <Disabled> MorphologyType: %s, ", morph_type or "<None>")
        return
    if morph_type is None:
        logging.warning("MorphologyType not set. Assuming ascii and legacy /ascii subdir")
        morph_type = "asc"
        morph_path = os.path.join(morph_path, "ascii")
        config_dict["MorphologyType"] = morph_type
        config_dict["MorphologyPath"] = morph_path
    assert morph_type in ("asc", "swc", "h5", "hoc"), "Invalid MorphologyType"
    log_verbose(" > MorphologyType = %s, src: %s", morph_type, morph_path)


def _validate_file_extension(path):
    if not path:
        return
    filepath = path.split(":")[0]  # for sonata, remove the edge_pop name appended to the file path
    if filepath.endswith(".sonata"):
        raise ConfigurationError("*.sonata files are no longer supported, "
                                 "please rename them to *.h5")


@SimConfig.validator
def _base_circuit(config: _SimConfig, run_conf):
    log_verbose("CIRCUIT (default): %s", run_conf.get("CircuitPath", "<DISABLED>"))
    config.base_circuit = _make_circuit_config(run_conf)


@SimConfig.validator
def _extra_circuits(config: _SimConfig, run_conf):
    from . import EngineBase
    extra_circuits = {}

    for name, circuit_info in config._blueconfig.Circuit.items():
        log_verbose("CIRCUIT %s (%s)", name, circuit_info.get("Engine", "(default)"))
        if "Engine" in circuit_info:
            # Replace name by actual engine
            circuit_info["Engine"] = EngineBase.get(circuit_info["Engine"])
        else:
            # Without custom engine, inherit base circuit infos
            for field in ("CircuitPath", "MorphologyPath", "MorphologyType",
                          "METypePath", "CellLibraryFile"):
                if field in config.base_circuit and field not in circuit_info:
                    log_verbose(" > Inheriting '%s' from base circuit", field)
                    circuit_info[field] = config.base_circuit[field]
        circuit_info.setdefault("nrnPath", False)
        extra_circuits[name] = _make_circuit_config(circuit_info, req_morphology=False)
        extra_circuits[name]._name = name

    # Sort so that iteration is deterministic
    config.extra_circuits = dict(sorted(
        extra_circuits.items(),
        key=lambda x: (x[1].Engine.CircuitPrecedence if x[1].Engine else 0, x[0])
    ))


@SimConfig.validator
def _global_parameters(config: _SimConfig, run_conf):
    from neuron import h
    config.celsius = run_conf.get("Celsius", 34)
    config.v_init = run_conf.get("V_Init", -65)
    config.extracellular_calcium = run_conf.get("ExtracellularCalcium")
    config.buffer_time = 25 * run_conf.get("FlushBufferScalar", 1)
    config.tstop = run_conf["Duration"]
    h.celsius = config.celsius
    h.set_v_init(config.v_init)
    h.tstop = config.tstop
    config.default_neuron_dt = h.dt
    h.dt = run_conf.get("Dt", h.dt)
    h.steps_per_ms = 1.0 / h.dt
    props = ("celsius", "v_init", "extracellular_calcium", "tstop", "buffer_time")
    log_verbose("Global params: %s",
                " | ".join(p + ": %s" % getattr(config, p) for p in props))
    if "CompartmentsPerSection" in run_conf:
        logging.warning("CompartmentsPerSection is currently not supported. "
                        "If needed, please request with a detailed usecase.")


@SimConfig.validator
def _set_simulator(config: _SimConfig, run_conf):
    user_config = config.cli_options
    simulator = run_conf.get("Simulator")

    if simulator is None:
        run_conf["Simulator"] = simulator = "NEURON"
    if simulator not in ("NEURON", "CORENEURON"):
        raise ConfigurationError("'Simulator' value must be either NEURON or CORENEURON")
    if simulator == "NEURON" and (user_config.build_model is False
                                  or user_config.simulate_model is False):
        raise ConfigurationError("Disabling model building or simulation is only"
                                 " compatible with CoreNEURON")

    log_verbose("Simulator = %s", simulator)
    config.use_neuron = (simulator == "NEURON")
    config.use_coreneuron = (simulator == "CORENEURON")


@SimConfig.validator
def _spike_parameters(config: _SimConfig, run_conf):
    spike_location = run_conf.get("SpikeLocation", "soma")
    if spike_location not in ["soma", "AIS"]:
        raise ConfigurationError("Possible options for SpikeLocation are 'soma' and 'AIS'")
    spike_threshold = run_conf.get("SpikeThreshold", -30)
    log_verbose("Spike_Location = %s", spike_location)
    log_verbose("Spike_Threshold = %s", spike_threshold)
    config.spike_location = spike_location
    config.spike_threshold = spike_threshold


_condition_checks = {
    "secondorder": (
        (0, 1, 2),
        ConfigurationError(
            "Time integration method (SecondOrder value) {} is invalid. Valid options are:"
            " '0' (implicitly backward euler),"
            " '1' (Crank-Nicolson) and"
            " '2' (Crank-Nicolson with fixed ion currents)")
    ),
    "randomize_Gaba_risetime": (
        ("True", "False", "0", "false"),
        ConfigurationError("randomize_Gaba_risetime must be True or False")
    ),
    "SYNAPSES__minis_single_vesicle": ((0, 1), None),
    "synapses__minis_single_vesicle": ((0, 1), None),
    "SYNAPSES__init_depleted": ((0, 1), None),
    "synapses__init_depleted": ((0, 1), None),
}


@SimConfig.validator
def _simulator_globals(config: _SimConfig, run_conf):
    if not hasattr(config._blueconfig, "Conditions"):
        return None
    from neuron import h
    # Hackish but some constants only live in the helper
    h.load_file("GABAABHelper.hoc")

    for group in config._blueconfig.Conditions.values():
        for key, value in group.items():
            validator = _condition_checks.get(key)
            if validator:
                config_exception = validator[1] or ConfigurationError(
                    "Value {} not valid for key {}. Allowed: {}".format(value, key, validator[0])
                )
                if value not in validator[0]:
                    raise config_exception
            synvar_prefix = "SYNAPSES__"
            if key.startswith(synvar_prefix) or key.startswith(synvar_prefix.lower()):
                key = key[len(synvar_prefix):]
                config.synapse_options[key] = value
                log_verbose("SYNAPSES %s = %s", key, value)
                for synapse_name in ("ProbAMPANMDA_EMS", "ProbGABAAB_EMS", "GluSynapse"):
                    setattr(h, key + "_" + synapse_name, value)
            else:
                log_verbose("GLOBAL %s = %s", key, value)
                setattr(h, key, value)
            if "cao_CR" in key and value != config.extracellular_calcium:
                logging.warning("Value of %s (%s) is not the same as extracellular_calcium (%s)"
                                % (key, value, config.extracellular_calcium))


@SimConfig.validator
def _second_order(config: _SimConfig, run_conf):
    second_order = run_conf.get("SecondOrder")
    if second_order is None:
        return None
    second_order = int(second_order)
    if second_order in (0, 1, 2):
        from neuron import h
        log_verbose("SecondOrder = %g", second_order)
        config.second_order = second_order
        h.secondorder = second_order
    else:
        raise _condition_checks["secondorder"][1]


@SimConfig.validator
def _single_vesicle(config: _SimConfig, run_conf):
    if "MinisSingleVesicle" not in run_conf:
        return
    from neuron import h
    if not hasattr(h, "minis_single_vesicle_ProbAMPANMDA_EMS"):
        raise NotImplementedError("Synapses don't implement minis_single_vesicle. "
                                  "More recent neurodamus model required.")
    minis_single_vesicle = int(run_conf["MinisSingleVesicle"])
    log_verbose("minis_single_vesicle = %d", minis_single_vesicle)
    h.minis_single_vesicle_ProbAMPANMDA_EMS = minis_single_vesicle
    h.minis_single_vesicle_ProbGABAAB_EMS = minis_single_vesicle
    h.minis_single_vesicle_GluSynapse = minis_single_vesicle


@SimConfig.validator
def _randomize_gaba_risetime(config: _SimConfig, run_conf):
    randomize_risetime = run_conf.get("RandomizeGabaRiseTime")
    if randomize_risetime is None:
        return
    from neuron import h
    h.load_file("GABAABHelper.hoc")
    if not hasattr(h, "randomize_Gaba_risetime"):
        raise NotImplementedError("Models don't support setting RandomizeGabaRiseTime. "
                                  "Please load a more recent model or drop the option.")
    assert randomize_risetime in ("True", "False", "0", "false")  # any non-"True" value is negative
    log_verbose("randomize_Gaba_risetime = %s", randomize_risetime)
    h.randomize_Gaba_risetime = randomize_risetime


@SimConfig.validator
def _current_dir(config: _SimConfig, run_conf):
    curdir = run_conf.get("CurrentDir")
    run_conf["BlueConfigDir"] = config.blueconfig_dir

    if curdir is None:
        log_verbose("CurrentDir using BlueConfig path [default]")
        curdir = config.blueconfig_dir
    else:
        if not os.path.isabs(curdir):
            if curdir == ".":
                logging.warning("Setting CurrentDir to '.' is discouraged and "
                                "shall never be used in production jobs.")
            else:
                raise ConfigurationError("CurrentDir: Relative paths not allowed")
            curdir = os.path.abspath(curdir)
        if not os.path.isdir(curdir):
            raise ConfigurationError("CurrentDir doesnt exist: " + curdir)

    log_verbose("CurrentDir = %s", curdir)
    run_conf["CurrentDir"] = curdir
    config.current_dir = curdir


@SimConfig.validator
def _output_root(config: _SimConfig, run_conf):
    def check_oputput_directory(output_dir):
        """
        Checks if output directory exists and is a directory.
        If it doesn't exists create it.
        This logic is based in old utility.mod.
        """
        if os.path.exists(output_dir):
            if not os.path.isdir(output_dir):
                raise Exception(f"{output_dir} does not name a directory.")
        else:
            try:
                os.makedirs(output_dir)
            except Exception as e:
                raise Exception(f"Failed to create OutputRoot directory {output_dir} with {e}")
            log_verbose(f"Directory {output_dir} does not exist.  Creating...")

    """confirm output_path exists and is usable"""
    output_path = run_conf.get("OutputRoot")

    if config.cli_options.output_path not in (None, output_path):
        output_path = config.cli_options.output_path
    if output_path is None:
        raise ConfigurationError("'OutputRoot' configuration not set")
    if not os.path.isabs(output_path):
        output_path = os.path.join(config.current_dir, output_path)

    from ._neurodamus import MPI
    if MPI.rank == 0:
        check_oputput_directory(output_path)
        # Delete coreneuron_input link since it may conflict with restore
        corenrn_input = output_path + "/coreneuron_input"
        if os.path.islink(corenrn_input):
            os.remove(corenrn_input)

    # Barrier to make sure that the output_path is created in case it doesn't exist
    MPI.barrier()

    log_verbose("OutputRoot = %s", output_path)
    run_conf["OutputRoot"] = output_path
    config.output_root = output_path


@SimConfig.validator
def _check_save(config: _SimConfig, run_conf):
    cli_args = config.cli_options
    save_path = cli_args.save or run_conf.get("Save")
    save_time = cli_args.save_time or run_conf.get("SaveTime")

    if not save_path:
        if save_time:
            logging.warning("SaveTime/--save-time IGNORED. Reason: no 'Save' config entry")
        return

    if not config.use_coreneuron:
        raise ConfigurationError("Save-restore only available with CoreNeuron")

    # Handle save
    assert isinstance(save_path, str), "Save must be a string path"
    if save_time:
        save_time = float(save_time)
        if save_time > config.tstop:
            logging.warning("SaveTime specified beyond Simulation Duration. "
                            "Setting SaveTime to simulation end time.")
            save_time = None

    config.save = os.path.join(config.current_dir, save_path)
    config.save_time = save_time


@SimConfig.validator
def _check_restore(config: _SimConfig, run_conf):
    restore = config.cli_options.restore or run_conf.get("Restore")
    if not restore:
        return

    if not config.use_coreneuron:
        raise ConfigurationError("Save-restore only available with CoreNeuron")

    # sync restore settings to hoc, otherwise we end up with an empty coreneuron_input dir
    run_conf["Restore"] = restore

    restore_path = os.path.join(config.current_dir, restore)
    assert os.path.isdir(os.path.dirname(restore_path))
    config.restore = restore_path


@SimConfig.validator
def _coreneuron_params(config: _SimConfig, run_conf):
    buffer_size = run_conf.get("ReportingBufferSize")
    if buffer_size is not None:
        assert buffer_size > 0
        log_verbose("ReportingBufferSize = %g", buffer_size)
        config.corenrn_buff_size = int(buffer_size)

    # Set defaults for CoreNeuron dirs since SimConfig init/verification happens after
    config.coreneuron_outputdir = config.output_root
    coreneuron_datadir = os.path.join(config.output_root, "coreneuron_input")

    if config.use_coreneuron and config.restore:
        # Most likely we will need to reuse coreneuron_input from first part
        if not os.path.isdir(coreneuron_datadir):
            logging.info("RESTORE: Searching for coreneuron_input besides " + config.restore)
            coreneuron_datadir = os.path.join(config.restore, "..", "coreneuron_input")
        assert os.path.isdir(coreneuron_datadir), "coreneuron_input dir not found"

    config.coreneuron_datadir = coreneuron_datadir


@SimConfig.validator
def _check_model_build_mode(config: _SimConfig, run_conf):
    user_config = config.cli_options
    config.build_model = user_config.build_model
    config.simulate_model = user_config.simulate_model

    if config.build_model is True:  # just create the data
        return
    if config.use_coreneuron is False:
        if config.build_model is False:
            raise ConfigurationError("Skipping model build is only available with CoreNeuron")
        else:  # was None
            config.build_model = True
            return

    # CoreNeuron restore is a bit special and already had its input dir checked
    if config.restore:
        config.build_model = False
        return

    # It's a CoreNeuron run. We have to check if build_model is AUTO or OFF
    core_data_location = config.coreneuron_datadir

    try:
        # Ensure that 'sim.conf' and 'files.dat' exist, and that '/dev/shm' was not used
        with open(os.path.join(config.output_root, "sim.conf"), 'r') as f:
            core_data_exists = (
                "datpath='/dev/shm/" not in f.read()
                and os.path.isfile(os.path.join(core_data_location, "files.dat"))
            )
    except FileNotFoundError:
        core_data_exists = False

    if config.build_model in (None, "AUTO"):
        # If enable-shm option is given we have to rebuild the model and delete any previous files
        # in /dev/shm or gpfs
        # In any case when we start model building any data in the core_data_location_shm if it
        # exists are deleted
        if not core_data_exists:
            core_data_location_shm = SHMUtil.get_datadir_shm(core_data_location)
            data_location = (
                core_data_location_shm
                if (
                    user_config.enable_shm and core_data_location_shm is not None
                )
                else core_data_location
            )
            logging.info("Valid CoreNeuron input data not found in '%s'. "
                         "Neurodamus will proceed to model building.",
                         data_location)
            config.build_model = True
        else:
            logging.info("CoreNeuron input data found in %s. Skipping model build",
                         core_data_location)
            config.build_model = False

    if not config.build_model and not core_data_exists:
        raise ConfigurationError("Model build DISABLED but no CoreNeuron data found")


@SimConfig.validator
def _keep_coreneuron_data(config: _SimConfig, run_conf):
    if config.use_coreneuron:
        keep_core_data = False
        if config.cli_options.keep_build or run_conf.get("KeepModelData", False) == "True":
            keep_core_data = True
        elif not config.cli_options.simulate_model or config.save:
            logging.warning("Keeping coreneuron data for CoreNeuron following run")
            keep_core_data = True
        config.delete_corenrn_data = not keep_core_data
    log_verbose("delete_corenrn_data = %s", config.delete_corenrn_data)


@SimConfig.validator
def _model_building_steps(config: _SimConfig, run_conf):
    user_config = config.cli_options
    if user_config.modelbuilding_steps is not None:
        ncycles = int(user_config.modelbuilding_steps)
        src_is_cli = True
    elif "ModelBuildingSteps" in run_conf:
        ncycles = int(run_conf["ModelBuildingSteps"])
        src_is_cli = False
    else:
        return None

    assert ncycles > 0, "ModelBuildingSteps set to 0. Required value > 0"

    if not SimConfig.use_coreneuron:
        logging.warning("IGNORING ModelBuildingSteps since simulator is not CORENEURON")
        return None

    if "CircuitTarget" not in run_conf:
        raise ConfigurationError(
            "Multi-iteration coreneuron data generation requires CircuitTarget")

    logging.info("Splitting Target for multi-iteration CoreNeuron data generation")
    logging.info(" -> Cycles: %d. [src: %s]", ncycles, "CLI" if src_is_cli else "BlueConfig")
    config.modelbuilding_steps = ncycles


@SimConfig.validator
def _report_vars(config: _SimConfig, run_conf):
    """Compartment reports read voltages or i_membrane only. Other types must be summation"""
    mandatory_fields = ("Type", "StartTime", "Target", "Dt", "ReportOn", "Unit", "Format")
    report_types = {"compartment", "Summation", "Synapse", "PointType", "lfp"}
    non_negatives = ("StartTime", "EndTime", "Dt")
    report_configs_dict = {}

    for rep_name, rep_config in config.reports.items():
        rep_config = compat.Map(rep_config).as_dict(parse_strings=True)
        report_configs_dict[rep_name] = rep_config

        _check_params("Report " + rep_name, rep_config, mandatory_fields,
                      non_negatives=non_negatives,
                      valid_values={'Type': report_types})

        if rep_config["Format"] != "SONATA":
            raise ConfigurationError(
                f"Unsupported report format: '{rep_config['Format']}'. Use 'SONATA' instead.")

        if config.use_coreneuron and rep_config["Type"] == "compartment":
            if rep_config["ReportOn"] not in ("v", "i_membrane"):
                logging.warning("Compartment reports on vars other than v and i_membrane "
                                " are still not fully supported (CoreNeuron)")
    # Overwrite config with a pure dict since we never need underlying hoc map
    config.reports = report_configs_dict


@SimConfig.validator
def _spikes_sort_order(config: _SimConfig, run_conf):
    order = run_conf.get("SpikesSortOrder", "by_time")
    if order not in ["none", "by_time"]:
        raise ConfigurationError("Unsupported spikes sort order %s, " % order +
                                 "BBP supports 'none' and 'by_time'")


def get_debug_cell_gid(cli_options):
    gid = None if not cli_options else cli_options.get("dump_cell_state")
    try:
        gid = int(gid) if gid is not None else SimConfig.run_conf.get("prCellGid")
    except ValueError as e:
        raise ConfigurationError("Cannot parse Gid for dump-cell-state: " + gid) from e
    if gid and SimConfig.is_sonata_config:
        # In sonata mode, user will provide a 0-based, add 1.
        gid += 1
    return gid


def check_connections_configure(SimConfig, target_manager):
    """Check connection block configuration and raise warnings for:
    1. Global variable should be set in the Conditions block,
    2. Connection overriding chains (t=0)
    3. Connections with Weight=0 not overridden by delayed (not instantiated)
    4. Partial Connection overriding -> Error
    5. Connections with delay > 0 not overriding anything
    """
    config_assignment = re.compile(r"(\S+)\s*\*?=\s*(\S+)")
    processed_conn_blocks = []
    zero_weight_conns = []
    conn_configure_global_vars = defaultdict(list)

    def get_overlapping_connection_pathway(base_conns, conn):
        for base_conn in reversed(base_conns):
            if target_manager.pathways_overlap(base_conn, conn):
                yield base_conn

    def process_t0_parameter_override(conn):
        if float(conn.get("Weight", 1)) == 0:
            zero_weight_conns.append(conn)
        for overridden_conn in get_overlapping_connection_pathway(processed_conn_blocks, conn):
            conn["_overrides"] = overridden_conn
            if target_manager.pathways_overlap(conn, overridden_conn, equal_only=True):
                overridden_conn["_full_overridden"] = True
            break  # We always compute only against the first overridden block
        processed_conn_blocks.append(conn)

    def process_weight0_override(conn):
        is_overriding = False
        for conn2 in get_overlapping_connection_pathway(zero_weight_conns, conn):
            is_overriding = True
            # If there isn't a full override for zero weights, we must raise exception (later)
            if not conn2.get("_full_overridden"):
                conn2["_full_overridden"] = target_manager.pathways_overlap(conn, conn2, True)
        if not is_overriding:
            logging.warning("Delayed connection %s is not overriding any weight=0 Connection",
                            conn["_name"])

    def get_syn_config_vars(conn):
        return [var for var, _ in config_assignment.findall(conn.get("SynapseConfigure", ""))]

    def display_overriding_chain(conn):
        logging.warning("Connection %s takes part in overriding chain:", conn["_name"])
        while conn is not None:
            logging.info(
                " -> %-6s %-60.60s Weight: %-8s SpontMinis: %-8s SynConfigure: %s",
                "(base)" if not conn.get("_overrides") else " ^",
                "{0[_name]}  {0[Source]} -> {0[Destination]}".format(conn),
                conn.get("Weight", "-"),
                conn.get("SpontMinis", "-"),
                ", ".join(get_syn_config_vars(conn))
            )
            if conn.get("_visited"):
                break
            conn["_visited"] = True
            conn = conn.get("_overrides")

    logging.info("Checking Connection Configurations")
    all_conn_blocks = [compat.Map(conn).as_dict(parse_strings=True)
                       for conn in SimConfig.connections.values()]

    # On a first phase process only for t=0
    for name, conn_conf in zip(SimConfig.connections, all_conn_blocks):
        conn_conf["_name"] = name
        if float(conn_conf.get("Delay", 0)) > .0:
            continue
        for var in get_syn_config_vars(conn_conf):
            if not var.startswith("%s"):
                conn_configure_global_vars[name].append(var)
        # Process all conns to show full override chains to help debug
        process_t0_parameter_override(conn_conf)

    # Second phase: find overridden zero_weights at t > 0
    for conn_conf in all_conn_blocks:
        if float(conn_conf.get("Delay", 0)) > 0:
            process_weight0_override(conn_conf)

    # CHECK 1: Global vars
    if conn_configure_global_vars:
        logging.warning("Global variables in SynapseConfigure. Review the following "
                        "connections and move the global vars to Conditions block")
        for name, vars in conn_configure_global_vars.items():
            logging.warning(" -> %s: %s", name, vars)
    else:
        logging.info(" => CHECK No Global vars!")

    # CHECK 2: Block override chains
    if not [display_overriding_chain(conn_conf)
            for conn_conf in reversed(processed_conn_blocks)
            if conn_conf.get("_overrides") and not conn_conf.get("_visited")]:
        logging.info(" => CHECK No Block Overrides!")

    # CHECK 3: Weight 0 not/badly overridden
    not_overridden_weight_0 = []
    for conn in zero_weight_conns:
        full_overriden = conn.get("_full_overridden")
        if full_overriden is None:
            not_overridden_weight_0.append(conn)
        elif full_overriden is False:  # incomplete override
            raise ConfigurationError(
                "Partial Weight=0 override is not supported: Conn %s" % conn["_name"])
    if not_overridden_weight_0:
        logging.warning("The following connections with Weight=0 are not overridden, "
                        "thus won't be instantiated:")
        for conn in not_overridden_weight_0:
            logging.warning(" -> %s", conn["_name"])
    else:
        logging.info(" => CHECK No single Weight=0 blocks!")


def _input_resistance(config: _SimConfig, target_manager):
    prop = "@dynamics:input_resistance"
    for stim_name, stim in config.stimuli.items():
        stim = compat.Map(stim).as_dict(parse_strings=True)
        stim_inject = config.get_stim_inject(stim_name)
        if stim_inject is None:
            continue  # not injected, do not care
        target_name = stim_inject["Target"]
        if stim["Mode"] == "Conductance" and \
           stim["Pattern"] in ["RelativeShotNoise", "RelativeOrnsteinUhlenbeck"]:
            # NOTE: use target_manager to read the population names of hoc or NodeSet targets
            target = target_manager.get_target(target_name)
            for population in target.population_names:
                config._cell_requirements.setdefault(population, set()).add(prop)
                log_verbose('[cell] %s (%s:%s)' % (prop, population, target.name))



"""
    timeit classes, functions, decorators and context managers

NOTE: Implementation is done along the lines of timeit.hoc

You can use the @timeit as decorator for functions
    >>> from neurodamus.utils.timeit import timeit
    >>> @timeit(name="test_func")
    >>> def test_func():
    >>>     from time import sleep
    >>>     sleep(1)

or as contextmanager for blocks of code
    >>> from neurodamus.utils.timeit import timeit
    >>> def test_block():
    >>>     with timeit(name="test_block"):
    >>>         from time import sleep
    >>>         sleep(2)

Similar to timeit.hoc, you can print every timer just after the measurement.
NOTE: the printing is effectively done only if you run neurodamus with <--verbose>.
For x-reference purposes, we print with "setpvec/accum" keywords:

    [ VERB ]  -> setpvec 15                  Cell creation 12.5351

The difference is that "accum" is printed from 2nd call on + sequence no. joined with "setpvec"

   [ VERB ]  -> accum 16                  Cell creation 10.7224 => TotalTime: 21.5789

One use case would be to see the time for each iteration of a loop.
This can be done by using the ctxt manager right after the for:
    >>> from neurodamus.utils.timeit import timeit
    >>> def test_block():
    >>>     for x in range():
    >>>         with timeit(name="test_block_in_for"):
    >>>             test_block_in_for()
    >>>             ....
    [ VERB ]  -> setpvec 1                test_block_in_for 1
    [ VERB ]  -> accum 2                  test_block_in_for 1 => TotalTime: 2
    [ VERB ]  -> accum 3                  test_block_in_for 2 => TotalTime: 4
    [ VERB ]  -> accum 4                  test_block_in_for 3 => TotalTime: 7
    [ VERB ]  -> accum 5                  test_block_in_for 4 => TotalTime: 11

You can however skip printing them, useful for big loops for example; you will in turn
get the total time at the end. Set <verbose> parameter to false for that:
    >>> from neurodamus.utils.timeit import timeit
    >>> def test_block():
    >>>     for x in big_range():
    >>>         ....
    >>>         with timeit(name="test_block_in_for", verbose = False):
    >>>             test_block_in_for()
    >>>         ....


There is a TimerManager singleton that holds all the timers.
You can call <timeit_show_stats()> on it to get statistics
    >>> from neurodamus.utils.timeit import TimerManager
    >>> TimerManager.timeit_show_stats()
    [ INFO ] +====================== TIMEIT STATS =============================+
    [ INFO ] |           Event Label          | Avg.Time | Min.Time | Max.Time |
    [ INFO ] +-----------------------------------------------------------------+
    [ INFO ] |                      test_func       1.00 |     1.00 |     1.00 |
    [ INFO ] |                     test_block       2.00 |     2.00 |     2.00 |
    [ INFO ] +-----------------------------------------------------------------+

You can archive timers, useful for multiple flow runs (see node.py MULTI-CYCLE RUN)
or just to break down numbers. You need to call <archive()> from TimerManager:
    >>> for index in range(2):
    >>>     test_func()
    >>>     test_block()
    >>>     TimerManager.archive(archive_name=" Test {:d}".format(index + 1))
    >>> TimerManager.timeit_show_stats()
    ....
    [ INFO ] +==================== TIMEIT STATS ( Test 1) =====================+
    [ INFO ] |           Event Label          | Avg.Time | Min.Time | Max.Time |
    [ INFO ] +-----------------------------------------------------------------+
    [ INFO ] |                      test_func       1.00 |     1.00 |     1.00 |
    [ INFO ] |                     test_block       2.00 |     2.00 |     2.00 |
    [ INFO ] +-----------------------------------------------------------------+

    [ INFO ] +==================== TIMEIT STATS ( Test 2) =====================+
    [ INFO ] |           Event Label          | Avg.Time | Min.Time | Max.Time |
    [ INFO ] +-----------------------------------------------------------------+
    [ INFO ] |                      test_func       1.00 |     1.00 |     1.00 |
    [ INFO ] |                     test_block       2.00 |     2.00 |     2.00 |
    [ INFO ] +-----------------------------------------------------------------+
    ....

In case you want to time operations just on rank0, without impacting the STATS,you can
use @timeit_rank0 in a similar manner to @timeit. For example delete_corenrn_data,
which is happening only on rank0.

PITFALLS:
    Make sure you avoid data-driven flows for your timeits, otherwise there will be
    an MPI_Allreduce number of timers mismatch and you will get MPI errors like:
        MPT: --------stack traceback-------
        MPT ERROR: rank:3426, function:MPI_ALLREDUCE, Message truncated on receive:
                        An application bug caused the sender to send too much data
    Example:
          for conn in self.get_target_connections(None, target_name):
            if conn.sgid not in spike_manager:
                continue
            # PITFALL: May not be called on some ranks => one less timer
            with timeit(name="register replay events", verbose=False):
                conn.replay(spike_manager[conn.sgid], start_delay)
                replayed_count += 1

"""
from __future__ import absolute_import
import logging
import time

from contextlib import contextmanager, ContextDecorator
from itertools import chain
from math import log, floor

from .logging import log_verbose
from ..core import NeurodamusCore as Nd, MPI, run_only_rank0


def human_readable(num):
    """ Get a human readable version of a number.
    For example: 5200000 -> 5.20M

    Args:
        num: The number
    """
    units = ['', 'K', 'M', 'B', 'T', 'P']
    power = int(floor(log(num, 1000.)))
    return '{:.2f}{:s}'.format(num / 1000. ** power, units[power]) if power >= 1 \
        else str(int(num))


delim = u'\u255a'


class _Timer(object):
    total_time = property(lambda self: self._total_time)
    accumulated = property(lambda self: self._accumulated)
    name = property(lambda self: self._name)
    hits = property(lambda self: self._hits)

    def __init__(self, name):
        self._name = name
        self._total_time = 0
        self._start_time = None
        self._last_time = None
        self._accumulated = False
        self._hits = 0

    def start(self):
        self._start_time = time.perf_counter()
        return self

    def stop(self):
        self._last_time = time.perf_counter() - self._start_time
        if self._total_time:
            self._accumulated = True
        self._total_time += self._last_time
        self._hits += 1
        self._start_time = None  # invalidate start time

    def log(self, keyword, seq_no=None):
        log_verbose("{:s} {} {:<30s} {:.4f} {:s}".
                    format(keyword,
                           seq_no if seq_no is not None else '',
                           self._name,
                           self._last_time,
                           "=> TotalTime: {:g}".format(self._total_time) if self._accumulated
                           else ""))


class _TimerManager(object):
    _timers = dict()
    _timers_sequence = 0
    _archived_timers = {}

    # archive current timers
    def archive(self, archive_name):
        self._archived_timers[archive_name] = self._timers
        self._timers = dict()

    def init(self, name):
        self._timers.setdefault(name, _Timer(name))
        self._timers[name].start()

    def update(self, name, verbose=True):
        if name not in self._timers.keys():
            raise Exception("{} not initialized in timers dict".format(name))
        self._timers[name].stop()
        if verbose:
            self._log_timer(self._timers[name])

    @run_only_rank0
    def _log_timer(self, timer_info):
        timer_info.log(keyword="accum" if timer_info.accumulated else "setpvec",
                       seq_no=self._timers_sequence)
        self._timers_sequence += 1

    # Note: method name kept for reference wrt neurodamus-core timeit.hoc
    def timeit_show_stats(self):
        current_timers_name = "Final Stats" if len(self._archived_timers) else ""
        for timers_name, timers in chain(self._archived_timers.items(),
                                         ((current_timers_name, self._timers),)):
            mpi_times = Nd.Vector(tinfo.total_time for tinfo in timers.values())
            avg_times = mpi_times.c()
            MPI.pc.allreduce(avg_times, MPI.SUM)
            min_times = mpi_times.c()
            MPI.pc.allreduce(min_times, MPI.MIN)
            max_times = mpi_times.c()
            MPI.pc.allreduce(max_times, MPI.MAX)

            nof_hits = Nd.Vector(tinfo.hits for tinfo in timers.values())
            MPI.pc.allreduce(nof_hits, MPI.SUM)

            self._log_stats(timers_name, timers, avg_times, min_times, max_times, nof_hits)

    @run_only_rank0
    def _log_stats(self, timers_name, timers, avg_times, min_times, max_times, nof_hits):
        stats_name = " TIMEIT STATS {}".format('(' + timers_name + ') ' if timers_name
                                               else timers_name)
        logging.info("+{:=^111s}+".format(stats_name))
        logging.info("|{:^58s}|{:^10s}|{:^10s}|{:^10s}|{:^19s}|".format(
            'Event Label', 'Avg.Time', 'Min.Time', 'Max.Time', 'Hits R0 / Total '))
        logging.info("+{:-^111s}+".format('-'))

        for t, (name, tinfo) in enumerate(timers.items()):
            base_name = delim.join('  ') * name.count(delim) + name.split(delim)[-1]
            logging.info("| {:<56s} | {:8.2f} | {:8.2f} | {:8.2f} | {:>7s} / {:<7s} |".format(
                base_name,
                avg_times.x[t] / MPI.size,
                min_times.x[t],
                max_times.x[t],
                human_readable(tinfo.hits),
                human_readable(nof_hits.x[t])))
        logging.info("+{:-^111s}+".format('-'))


TimerManager = _TimerManager()  # singleton


# Can be used as context manager or decorator
class timeit(ContextDecorator):
    curr_path = []

    def __init__(self, name, verbose=True):
        self._original_name = name
        self._verbose = verbose

    def __enter__(self):
        timeit.curr_path.append(self._original_name)
        self._name = delim.join(timeit.curr_path)
        TimerManager.init(self._name)

    def __exit__(self, exc_type, exc, exc_tb):
        TimerManager.update(self._name, self._verbose)
        timeit.curr_path.pop()


# Can also be used as decorator.
# Does not use the TimerManager.
@contextmanager
def timeit_rank0(name):
    if MPI.rank == 0:
        t = _Timer(name).start()
        yield
        t.stop()
        t.log(keyword="rank0")
    else:
        yield  # No-op



# flake8: noqa
"""
    neurodamus
    ----------

    The neurodamus package implements the instantiation of simulations in Neuron
    based on a configuration file, a.k.a. BlueConfigs
    It is deeply based on the HOC implementation, therefore providing python modules like
    `node`, `cell_distributor`, etc; and still depends on several low-level HOC files at runtime.
"""
from __future__ import absolute_import
import pkg_resources
try:
    __version__ = pkg_resources.get_distribution(__name__).version
except Exception:
    __version__ = 'devel'

__author__ = "Fernando Pereira <fernando.pereira@epfl.ch>"
__copyright__ = "2018 Blue Brain Project, EPFL"

# Neurodamus node for setting up a neurodamus execution
from .node import Node, Neurodamus


__all__ = ["Node", "Neurodamus"]



"""
A wrapper for MorphIO objects. Provides additional neuron
features on top of MorphIO basic morphology handling.
"""
import os
import logging
import numpy as np
from numpy.linalg import eig, norm


'''
    [START] Implementations retrieved from nse/morph-tool (!= hpc/morpho-tool !!!)
    These functions are needed for soma points computation (a la import3d).
    TODO: Once we have something stable, integrate nse/morph-tool
'''
X, Y, Z, R = 0, 1, 2, 3


def split_morphology_path(morphology_path):
    '''Split `{collection_path}/{morph_name}.{ext}`'''

    if os.path.exists(morphology_path):
        collection_path = os.path.dirname(morphology_path)
        morph_name, morph_ext = os.path.splitext(os.path.basename(morphology_path))
        return collection_path, morph_name, morph_ext

    else:
        collection_path = morphology_path
        while not os.path.exists(collection_path):
            assert collection_path != os.path.dirname(collection_path), "Failed to split path."
            collection_path = os.path.dirname(collection_path)

        morph_name, morph_ext = os.path.splitext(os.path.relpath(morphology_path, collection_path))

        return collection_path, morph_name, morph_ext


def contourcenter(xyz):
    '''python implementation of NEURON code: lib/hoc/import3d/import3d_sec.hoc '''
    # xyz = xyz.astype('d')
    # print(xyz.dtype)
    # print(xyz[0])
    POINTS = 101

    points = np.vstack((np.diff(xyz[:, [X, Y]], axis=0), xyz[0, [X, Y]]))
    perim = np.cumsum(np.hstack(((0,), norm(points, axis=1))))[:-1]

    d = np.linspace(0, perim[-1], POINTS)
    new_xyz = np.zeros((POINTS, 3))
    for i in range(3):
        new_xyz[:, i] = np.interp(x=d, xp=perim, fp=xyz[:, i])

    mean = np.mean(new_xyz, axis=0)

    return mean, new_xyz


def get_sides(points, major, minor):
    '''
    Circular permutation of the points so that the point with the largest
    coordinate along the major axis becomes the last point
    tobj = major.c.mul(d.x[i])  ###### uneeded? line 1191
    '''

    major_coord, minor_coord = np.dot(points, major), np.dot(points, minor)

    imax = np.argmax(major_coord)
    # pylint: disable=invalid-unary-operand-type
    major_coord, minor_coord = (np.roll(major_coord, -imax),
                                np.roll(minor_coord, -imax))

    imin = np.argmin(major_coord)

    sides = [major_coord[:imin][::-1], major_coord[imin:]]
    rads = [minor_coord[:imin][::-1], minor_coord[imin:]]
    return sides, rads


def make_convex(sides, rads):
    '''Keep only points that make path convex'''

    def convex_idx(m):
        '''Return index to elements of m that make it convex

        Note: not efficient at the moment
        # now we have the two sides without the min and max points (rads[0]=0)
        # we hope both sides now monotonically increase, i.e. convex
        # make it convex

        '''
        idx = np.ones_like(m, dtype=bool)
        last_val = m[-1]
        for i in range(len(m) - 2, -1, -1):
            if m[i] < last_val:
                last_val = m[i]
            else:
                idx[i] = False
        return idx

    for i_side in [0, 1]:
        ci = convex_idx(sides[i_side])
        sides[i_side] = sides[i_side][ci]
        rads[i_side] = rads[i_side][ci]
    return sides, rads


def contour2centroid(mean, points):
    '''this follows the function in
            lib/hoc/import3d/import3d_gui.hoc
       most of the comments are from there, so if you want to follow along, it should
       break up the function the same way
    '''
    logging.debug('Converting soma contour into a stack of cylinders')

    # find the major axis of the ellipsoid that best fits the shape
    # assuming (falsely in general) that the center is the mean

    points = (points - mean)
    eigen_values, eigen_vectors = eig(np.dot(points.T, points))

    # To be consistent with NEURON eigen vector directions
    eigen_vectors *= -1

    idx = np.argmax(eigen_values)
    major = eigen_vectors[:, idx]
    # minor is normal and in xy plane
    idx = 3 - np.argmin(eigen_values) - np.argmax(eigen_values)
    minor = eigen_vectors[:, idx]
    minor[2] = 0

    sides, rads = get_sides(points, major, minor)
    sides, rads = make_convex(sides, rads)

    tobj = np.sort(np.hstack(sides))
    new_major_coord = np.linspace(tobj[1], tobj[-2], 21)
    rads[0] = np.interp(new_major_coord, sides[0], rads[0])
    rads[1] = np.interp(new_major_coord, sides[1], rads[1])

    points = major * new_major_coord[:, np.newaxis] + mean
    diameters = np.abs(rads[0] - rads[1])

    # avoid 0 diameter ends
    diameters[0] = np.mean(diameters[:2])
    diameters[-1] = np.mean(diameters[-2:])

    return points, diameters


def _to_sphere(neuron):
    '''Convert a 3-pts cylinder or a 1-pt sphere into a circular
    contour that represents the same sphere'''
    radius = neuron.soma.diameters[0] / 2.
    N = 20
    points = np.zeros((N, 3))
    phase = 2 * np.pi / (N - 1) * np.arange(N)
    points[:, 0] = radius * np.cos(phase)
    points[:, 1] = radius * np.sin(phase)
    points += neuron.soma.points[0]
    neuron.soma.points = points
    neuron.soma.diameters = np.repeat(radius, N)


def single_point_sphere_to_circular_contour(neuron):
    '''Transform a single point soma that represents a sphere
    into a circular contour that represents the same sphere'''
    logging.debug('Converting 1-point soma (sperical soma) to circular contour '
                  'representing the same sphere')
    _to_sphere(neuron)


'''
    [END] Implementations retrieved from nse/morph-tool (!= hpc/morpho-tool !!!)
'''


class MorphIOWrapper:
    """
        A class that wraps a MorphIO object and gets everything ready for HOC usage
    """
    morph = property(lambda self: self._morph)
    section_index2name_dict = property(lambda self: self._sec_idx2names)
    section_typeid_distrib = property(lambda self: self._sec_typeid_distrib)

    def __init__(self, input_file, options=0):
        self._collection_dir, self._morph_name, self._morph_ext = split_morphology_path(input_file)
        self._options = options
        self._build_morph()
        self._sec_idx2names = {}
        self._build_sec_idx2names()
        self._build_sec_typeid_distrib()

    def _build_morph(self):
        """ Build immutable morphology, going trough mutable and applying neuron adjustemnts """
        try:
            # Lazy import morphio since it has an issue with execl
            from morphio import SomaType, Option
            from morphio import Morphology
            from morphio import Collection
        except ImportError as e:
            raise RuntimeError("MorphIO is not available") from e

        collection = Collection(self._collection_dir, extensions=[self._morph_ext])
        options = self._options | Option.nrn_order
        self._morph = collection.load(self._morph_name, options, mutable=True)

        # Re-compute the soma points as they are computed in import3d_gui.hoc
        if self._morph.soma_type not in {SomaType.SOMA_SINGLE_POINT, SomaType.SOMA_SIMPLE_CONTOUR}:
            raise Exception(
                'A H5 file morphology is not supposed to have a soma of type: {}'.format(
                    self._morph.soma_type))
        logging.debug(
            "(%s, %s, %s) has soma type : %s",
            self._collection_dir, self._morph_name, self._morph_ext, self._morph.soma_type
        )

        if self._morph.soma_type == SomaType.SOMA_SINGLE_POINT:
            ''' See NRN import3d_gui.hoc -> instantiate()
                            -> sphere_rep(xx, yy, zz, dd) '''
            single_point_sphere_to_circular_contour(self._morph)
        elif self._morph.soma_type == SomaType.SOMA_SIMPLE_CONTOUR:
            ''' See NRN import3d_gui.hoc -> instantiate()
                            -> contour2centroid(xx, yy, zz, dd, sec) '''
            mean, new_xyz = contourcenter(self._morph.soma.points)
            self._morph.soma.points, self._morph.soma.diameters = contour2centroid(
                mean, new_xyz)

        self._morph = Morphology(self._morph)

    def _build_sec_idx2names(self):
        """ Build section index to nrn section names mapping on top of MorphIO section.id """

        # soma is not accounted for in sections
        self._sec_idx2names[0] = 'soma'

        idx_adjust = 0  # section ids don't restart from 0 for each type in MorphIO
        current_type = -1  # used for index adjustment
        for i, sec in enumerate(self._morph.sections):
            index = i + 1
            # When we have a new type, update idx_adjust
            if self._morph.section_types[sec.id] != current_type:
                current_type = self._morph.section_types[sec.id]
                idx_adjust = i

            self._sec_idx2names[index] = self.name(current_type, i - idx_adjust)

    def _build_sec_typeid_distrib(self):
        """ Build typeid distribution on top of MorphIO section_types """

        '''
            This will hold np.array that will map the different type ids to the
            start id wrt to section_types and their count. For example, axon type(2)
            starts at section.id 0 and and totals 2724 sections.

                | type_id | start_id  |   count   |
                | ------- | --------- | --------- |
                |      2  |        0  |     2724  |
                |      3  |     2724  |       75  |


            >>> _sec_typeid_distrib
            array([[(2,    0, 2724)],
                   [(3, 2724,   75)]],
                  dtype=[('type_id', '<i8'), ('start_id', '<i8'), ('count', '<i8')])

            Then use it like this:
            >>> _sec_typeid_distrib[['type_id', 'start_id']]
            array([[(2,    0)],
                   [(3, 2724)]],
                  dtype={'names':['type_id','start_id'],....}
        '''
        self._sec_typeid_distrib = np.dstack(np.unique(self._morph.section_types,
                                                       return_counts=True,
                                                       return_index=True))[0]
        self._sec_typeid_distrib = np.concatenate(([(1, -1, 1)], self._sec_typeid_distrib), axis=0)
        self._sec_typeid_distrib.dtype = [('type_id', '<i8'), ('start_id', '<i8'), ('count', '<i8')]

    def morph_as_hoc(self):
        """ Uses morphio object to read and generate hoc commands just like import3d_gui.hoc """

        cmds = []

        '''
            We need to get the number of sections for each type in order
            to generate the create commands. E.g.:
                ( soma , 1  )
                ( dend , 52 )
                ( axon , 23 )
                ( apic , 5  )
        '''
        # generate create commands
        for [(type_id, count)] in self._sec_typeid_distrib[['type_id', 'count']]:
            tstr = self.type2name(type_id)
            tstr1 = "create {}[{}]".format(tstr, count)
            cmds.append(tstr1)
            tstr1 = self.mksubset(type_id, count, tstr)
            cmds.append(tstr1)

        cmds.append("forall all.append")

        # generate 3D soma points commands. Order is reversed wrt NEURON's soma points.
        cmds.extend(("soma {{ pt3dadd({:.8g}, {:.8g}, {:.8g}, {:.8g}) }}"
                     .format(p[0], p[1], p[2], d)
                     for p, d in zip(reversed(self._morph.soma.points),
                                     reversed(self._morph.soma.diameters))))

        # generate sections connect + their respective 3D points commands
        for i, sec in enumerate(self._morph.sections):
            index = i + 1
            tstr = self._sec_idx2names[index]

            if not sec.is_root:
                if sec.parent is not None:
                    parent_index = sec.parent.id + 1
                    tstr1 = self._sec_idx2names[parent_index]
                    tstr1 = "{} connect {}(0), {}".format(tstr1, tstr, 1)
                    cmds.append(tstr1)
            else:
                tstr1 = "soma connect {}(0), {}".format(tstr, 0.5)
                cmds.append(tstr1)

                # pt3dstyle does not impact simulation numbers. This will be kept for x-reference.
                # tstr1 = "{} {{ pt3dstyle(1, {:.8g}, {:.8g}, {:.8g}) }}".format
                #                   (tstr, mean[0], mean[1], mean[2])
                # cmds.append(tstr1)

            # 3D point info
            cmds.extend("{} {{ pt3dadd({:.8g}, {:.8g}, {:.8g}, {:.8g}) }}"
                        .format(tstr, p[0], p[1], p[2], d)
                        for p, d in zip(sec.points, sec.diameters))

        return cmds

    '''
         [START] Python versions of import3d_gui.hoc helper functions
         Note: nrn function names will be kept for reference
    '''

    _type2name_dict = {1: "soma",
                       2: "axon",
                       3: "dend",
                       4: "apic"}

    @classmethod
    def type2name(cls, type_id):
        """
        :param type_id: id of section type
        :return: name representation of the section type.
                 If not found in _type2name_dict, then default is:
                    if (type < 0): "minus_{}".format(-type)
                    else: "dend_{}".format(type)
        """
        return cls._type2name_dict.get(type_id) or \
            ("minus_{}".format(-type_id) if type_id < 0
             else "dend_{}".format(type_id))

    _mksubset_dict = {1: "somatic",
                      2: "axonal",
                      3: "basal",
                      4: "apical"}

    @classmethod
    def mksubset(cls, type_id, freq, type_name):
        """
        :param type_id: id of section type
        :param freq: frequency of that section type id
        :param type_name: the name of the section type
        :return: command to append section type to subset
        """
        tstr = cls._mksubset_dict.get(type_id) or \
            ("minus_{}set".format(-type_id) if type_id < 0
             else "dendritic_{}".format(type_id))

        tstr1 = "forsec \"{}\" {}.append".format(type_name, tstr)
        return tstr1

    @classmethod
    def name(cls, type_id, index):
        return "{}[{}]".format(cls.type2name(type_id), index)

    '''
        [END] Python versions of import3d_gui.hoc helper functions
    '''



import logging
import numpy

from .core import NeurodamusCore as Nd
from .core.configuration import ConfigurationError


class LFPManager:
    """
    Class handling the lfp functionality
    """
    def __init__(self):
        self._lfp_file = None

    def load_lfp_config(self, lfp_weights_file, population_list):
        """Loads lfp weigths from h5 file
        """
        logging.info("Reading LFP configuration info from '%s'", lfp_weights_file)
        import h5py
        try:
            self._lfp_file = h5py.File(lfp_weights_file, 'r')
        except IOError as e:
            raise ConfigurationError(f"Error opening LFP electrodes file: {e}")

        # Check that the file contains the required groups for at least 1 population
        populations_found = []
        for pop_name in population_list:

            req_groups = [f'/electrodes/{pop_name}/scaling_factors', f'{pop_name}/node_ids',
                          f'{pop_name}/offsets']
            if all(group in self._lfp_file for group in req_groups):
                populations_found.append(pop_name)

        if not populations_found:
            raise ConfigurationError("The LFP weights file does not contain the necessary datasets "
                                     "'scaling_factors', 'node_ids' and 'offsets' "
                                     "in any of the populations {}.".format(population_list))

    def get_sonata_node_id(self, gid, population_info):
        return population_info[0], gid - population_info[1] - 1

    def get_node_id_subsets(self, node_id, population_name):
        node_ids = self._lfp_file[population_name]["node_ids"]
        # Look for the index of the node_id
        index = numpy.where(numpy.array(node_ids) == node_id)[0][0]
        offsets_dataset = self._lfp_file[population_name]["offsets"]
        electrodes_dataset = self._lfp_file["electrodes"][population_name]["scaling_factors"]
        index_low = offsets_dataset[index]
        index_high = offsets_dataset[index+1]
        # Get the subset data for the node_id and the section index
        subset_data = electrodes_dataset[index_low:index_high, :]
        return subset_data

    def read_lfp_factors(self, gid, population_info=("default", 0)):
        """
        Reads the local field potential (LFP) factors for a specific gid
        from an HDF5 file and returns the factors as a Nd.Vector.

        Args:
        gid (int): The unique cell identifier
        population_info (Pair(str, int)): Population info ("population_name", population_offset)

        Returns:
        Nd.Vector: A vector containing the LFP factors for the specified gid
        """
        scalar_factors = Nd.Vector()
        if self._lfp_file:
            try:
                population_name, node_id = self.get_sonata_node_id(gid, population_info)
                subset_data = self.get_node_id_subsets(node_id, population_name)
                for electrode_factors in subset_data:
                    scalar_factors.append(Nd.Vector(electrode_factors))
            except (KeyError, IndexError) as e:
                logging.warning("Node id {} not found in the electrodes file for population {}: {}"
                                .format(node_id, population_name, str(e)))
        return scalar_factors

    def get_number_electrodes(self, gid, population_info=("default", 0)):
        """Get number of electrodes of a certain gid
        """
        num_electrodes = 0
        if self._lfp_file:
            try:
                population_name, node_id = self.get_sonata_node_id(gid, population_info)
                subset_data = self.get_node_id_subsets(node_id, population_name)
                num_electrodes = subset_data.shape[1]
            except (KeyError, IndexError) as e:
                logging.warning("Node id {} not found in the electrodes file for population {}: {}"
                                .format(node_id, population_name, str(e)))
        return num_electrodes



"""
Structures holding the several instantiated objects from the configuration
"""
import numpy
import logging
from os import path as Path
from .core import NeurodamusCore as Nd


class ElectrodeManager(object):
    """ Electrode Manager.
        Create and handle electrodes according to configuration
    """
    __slots__ = ['_pos', '_names', '_electrodes', '_count']

    def __init__(self, elec_path, elecs_conf):
        """Init for ElectrodeManager.
        Args:
            elec_path: (string) The path where te find the electrodes config
            elecs_conf: A python Mapping of the configurations
        """
        self._names = []
        self._electrodes = []
        self._count = len(elecs_conf)
        self._pos = numpy.empty((self._count, 3), dtype='int32')  # x,y,z
        cur_i = 0

        for name, conf in elecs_conf.items():
            el_file = conf.get("File")
            if not el_file:
                logging.warning("No File speficied for electrode %s. Skipping", name)
                continue
            el_filepath = Path.join(elec_path, el_file.s)

            # Get electrode, checking versions
            v = conf.get("Version")  # Either None or hocObject
            if v is not None and v.s == "5":
                tmp_elec = Nd.lookupTableV2(el_filepath)
                v = tmp_elec.vInfo()
            if v != 5:
                logging.warning("Invalid Electrode version found. Only version 5 is supported")
                continue

            # Get & convert xyz - int() base 0 is autodetect (eq to %i)
            xyz = [int(conf.get(vname).s, 0) for vname in "xyz"]

            # Store info
            self._names.append(name)
            self._electrodes.append(tmp_elec)
            self._pos[cur_i] = xyz
            cur_i += 1

    def get(self, elec_id):
        """Retrieves an electrode object given its name or index.

        Raises: KeyError if the element doesnt exist
        """
        if isinstance(elec_id, str):
            for i, cur_name in enumerate(self._names):
                if cur_name == elec_id:
                    elec_id = i
                    break
            else:
                raise KeyError("ElectrodeManager> Non-existing name: %s" % elec_id)
        return self._electrodes[elec_id]

    def find_near(self, x, y, z):
        """Retrieves an electrode object given a <x, y, z> position.

        Returns: the first Electrode object found within 5 micron (in each direction),
            or None if not found
        """
        all_dist = numpy.absolute(self._pos - (x, y, z))
        for elec, dist in zip(self._electrodes, all_dist):
            if (dist < 5).all():
                return elec
        return None

    def get_name(self, i):
        return self._names[i]

    def get_position(self, i):
        return self._pos[i]

    def __len__(self):
        return self._count



"""
Stimulus implementation where incoming synaptic events are replayed for a single gid
"""
from __future__ import absolute_import
import os
import logging
import numpy
from .utils.logging import log_verbose
from .utils.multimap import GroupedMultiMap
from .utils.timeit import timeit


class SpikeManager:
    """ Holds and manages gid spike time information, specially for Replay.

    A SynapseReplay stim can be used for a single gid that has all the synapses instantiated.
    Given an out.dat file from a previous run, this object uses a NetStim object to retrigger
    the synapses at the appropriate time as though the presynaptic cells were present and active.

    Internally the spikes are stored in a :py:class:`neurodamus.utils.multimap.GroupedMultiMap`
    """
    _ascii_spike_dtype = [('time', 'double'), ('gid', 'uint32')]

    @timeit(name="Replay init")
    def __init__(self, spike_filename, delay=0, population=None):
        """Constructor for SynapseReplay.

        Args:
            spike_filename: path to spike out file.
                if ext is .bin, interpret as binary file; otherwise, interpret as ascii
            delay: delay to apply to spike times
        """
        self._gid_fire_events = None
        # Nd.distributedSpikes = 0  # Wonder the effects of this
        self.open_spike_file(spike_filename, delay, population)

    #
    def open_spike_file(self, filename, delay, population=None):
        """Opens a given spike file.

        Args:
            filename: path to spike out file. Interpret as binary or ascii according to extension
            delay: delay to apply to spike times
        """
        # determine if we have binary or ascii file
        # TODO: filename should be able to handle relative paths,
        # using the Run.CurrentDir as an initial path
        # _read_spikes_xxx shall return numpy arrays
        if filename.endswith(".h5"):
            tvec, gidvec = self._read_spikes_sonata(filename, population)
        elif filename.endswith(".bin"):
            tvec, gidvec = self._read_spikes_binary(filename)
        else:
            tvec, gidvec = self._read_spikes_ascii(filename)

        if delay:
            tvec += delay

        self._store_events(tvec, gidvec)

    @classmethod
    def _read_spikes_sonata(cls, filename, population):
        import libsonata
        spikes_file = libsonata.SpikeReader(filename)
        if population not in spikes_file.get_population_names():
            raise MissingSpikesPopulationError("Spikes population not found: " + population)
        spikes = spikes_file[population]
        spike_dict = spikes.get_dict()
        return spike_dict["timestamps"], spike_dict["node_ids"] + 1

    @classmethod
    def _read_spikes_ascii(cls, filename):
        log_verbose("Reading ascii spike file %s", filename)
        # first line is '/scatter'
        spikes = numpy.loadtxt(filename, dtype=cls._ascii_spike_dtype, skiprows=1, ndmin=1)

        if len(spikes) > 0:
            log_verbose("Loaded %d spikes", len(spikes))
        else:
            logging.warning("No spike/gid found in spike file %s", filename)

        return spikes["time"], spikes["gid"]

    @staticmethod
    def _read_spikes_binary(filename):
        """Read in the binary file with spike events.

        Format notes: The first half of file is interpreted as double precision time values
        followed by an equal number of double precision gid values.
        File must be produced on the same architecture where NEURON will run (i.e. no byte-swapping)
        Read the data on the root node, broadcasting info - This is fine as long as the entire data
        set fits in a single node's memory. Should it become significantly larger, this could be
        replaced with a distributed model where each node loads a portion of the file, and exchanges
        with other nodes so as to ultimately hold data for local gids.
        """
        log_verbose("Reading Binary spike file %s", filename)
        # there *should* be a number of doubles (8 bytes) such that
        # it is divisible by 2 (half for time values, half for gids)
        statinfo = os.stat(filename)
        filesize = statinfo.st_size
        n_events = filesize // 16
        if not filesize % 16:
            logging.warning("File size doesn't conform to have same number of gids and times")

        with open(filename, "rb") as reader:
            tvec = numpy.fromfile(reader, "d", n_events)
            gidvec = numpy.fromfile(reader, "d", n_events).astype("uint32")

        log_verbose("Replay: Loaded %d spikes", len(tvec))

        return tvec, gidvec

    #
    def _store_events(self, tvec, gidvec):
        """Stores the events in the _gid_fire_events GroupedMultiMap.

        tvec and gidvec arguments should be numpy arrays
        """
        spike_map = GroupedMultiMap(gidvec, tvec)
        if self._gid_fire_events is None:
            self._gid_fire_events = spike_map
        else:
            self._gid_fire_events += spike_map

    def __len__(self):
        return len(self._gid_fire_events)

    def __getitem__(self, gid):
        return self._gid_fire_events.get(gid)

    def __contains__(self, gid):
        return gid in self._gid_fire_events

    def get_map(self):
        """Returns the :py:class:`GroupedMultiMap` with all the spikes."""
        return self._gid_fire_events

    def filter_map(self, pre_gids):
        """Returns a raw dict of pre_gid->spikes for the given pre gids."""
        return {key: self._gid_fire_events[key] for key in pre_gids}

    def dump_ascii(self, f, gid_offset=None):
        """Writes the spikes out, in compat ascii format.

        Args:
            f: The file name or handle
        """
        gids, times = self._gid_fire_events.flatten().data()
        if gid_offset:
            log_verbose("dump_ascii: add offset %d to gids", gid_offset)
            gids += gid_offset
        expanded_ds = numpy.stack((times, gids), axis=-1)

        if isinstance(f, str):
            # If given a filename we assume a new file is wanted, with new header
            with open(f, "w") as fx:
                fx.write("/scatter\n")
                numpy.savetxt(fx, expanded_ds, fmt='%.3lf\t%d')
        else:
            # If given a file handle, user wants control so we directly dump
            numpy.savetxt(f, expanded_ds, fmt='%.3lf\t%d')

        log_verbose("Replay: Written %d entries", len(expanded_ds))


class MissingSpikesPopulationError(Exception):
    """An exception triggered when a given node population is not found, we may want to handle"""
    pass



#!/usr/bin/env python
"""
A few ways to create a cell with two branches, one of then two sections long
"""
from neurodamus.core import Cell


def test_create_cell():
    builder = Cell.Builder
    c = (builder
         .add_soma(1)
         .add_dendrite("dend1", 2, 5)
         .attach(builder.DendriteSection("dend2", 3, 2).add("sub2_dend", 4, 2))
         .create())
    return c


def test_create_cell_2():
    c = (Cell.Builder
         .add_soma(1)
         .add_dendrite("dend1", 2, 5)
         # Three axon sections in line
         .append_axon("ax1", 3, 2).append("ax1_2", 4, 2).append("ax1_3", 3, 3)
         .create())
    return c


def test_create_cell_3():
    Dend = Cell.Builder.DendriteSection
    c = (Cell.Builder
         .add_soma(1)
         .add_dendrite("dend1", 2, 5)
         .attach(Dend("dend2", 3, 2)
                 .append("sub2_dend", 4, 2)
                 .get_root())
         .create())
    return c


if __name__ == "__main__":
    test_create_cell()
    test_create_cell_2()
    test_create_cell_3()
    Cell.show_topology()



"""
Neurodamus is a software for handling neuronal simulation using neuron.

Copyright (c) 2018 Blue Brain Project, EPFL.
All rights reserved
"""
from neurodamus import Neurodamus

if __name__ == "__main__":
    Neurodamus("BlueConfig", enable_reports=False, logging_level=3).dump_circuit_config()
